from __future__ import annotations

import json as jsonlib
import textwrap

from dissect.target import plugin
from dissect.target.helpers.docs import INDENT_STEP, get_plugin_overview
from dissect.target.plugin import Plugin, arg, export
from dissect.target.plugins.os.default._os import DefaultOSPlugin


def generate_functions_overview(
    functions: list[plugin.FunctionDescriptor] | None = None, include_docs: bool = False
) -> list[str]:
    """Generate a tree list of functions with optional documentation."""

    categorized_plugins = _categorize_functions(functions)
    plugin_descriptions = _generate_plugin_tree_overview(categorized_plugins, include_docs)

    plugins_list = textwrap.indent(
        "\n".join(plugin_descriptions) if plugin_descriptions else "None",
        prefix=INDENT_STEP,
    )

    failed_descriptions = []
    failed_items = plugin.failed()
    for failed_item in failed_items:
        module = failed_item.module
        exception = failed_item.stacktrace[-1].rstrip()
        failed_descriptions.append(
            textwrap.dedent(
                f"""\
                    Module: {module}
                    Reason: {exception}
                """
            )
        )

    failed_list = textwrap.indent(
        "\n".join(failed_descriptions) if failed_descriptions else "None\n",
        prefix=INDENT_STEP,
    )

    lines = [
        "Available plugins:",
        plugins_list,
        "",
        "Failed to load:",
        failed_list,
    ]
    return "\n".join(lines)


def generate_functions_json(functions: list[plugin.FunctionDescriptor] | None = None) -> str:
    """Generate a JSON representation of all available functions."""

    loaded = []
    failed = []

    for desc in functions or _get_default_functions():
        docstring = desc.func.__doc__.split("\n\n", 1)[0].strip() if desc.func.__doc__ else None
        arguments = []

        for name, _arg in desc.args:
            is_bool_action = _arg.get("action", "") in ("store_true", "store_false")
            arg_desc = {
                "name": name[0],
                # infer the type either by store_*, type argument and fallback to default str.
                # See: https://docs.python.org/3/library/argparse.html#type
                "type": "bool" if is_bool_action else getattr(_arg.get("type"), "__name__", "str"),
                "help": _arg.get("help"),
                "default": _arg.get("action") == "store_false"
                if is_bool_action
                else (_arg.get("default") or _arg.get("const")),
                # required can either be set explicitly or is implied with '--' style arguments.
                "required": _arg.get("required", False),
            }

            arguments.append(arg_desc)

        loaded.append(
            {
                "name": desc.name,
                "output": desc.output,
                "description": docstring,
                "arguments": arguments,
                "path": desc.path,
                "alias": desc.alias,
            }
        )

    if failures := plugin.failed():
        failed = [{"module": f.module, "stacktrace": "".join(f.stacktrace)} for f in failures]

    return jsonlib.dumps({"loaded": loaded, "failed": failed})


def _get_os_functions() -> list[plugin.FunctionDescriptor]:
    return [f for f in plugin.functions(index="__os__") if f.exported and f.module == DefaultOSPlugin.__module__]


def _get_default_functions() -> list[plugin.FunctionDescriptor]:
    return [f for f in plugin.functions() if f.exported] + _get_os_functions()


def _categorize_functions(functions: list[plugin.FunctionDescriptor] | None = None) -> dict:
    """Categorize functions based on its module path."""

    functions = functions or _get_default_functions()
    result = {}

    for desc in functions:
        obj = result
        parts = desc.path.split(".")

        if not desc.namespace or (desc.namespace and desc.method_name != "__call__"):
            parts = parts[:-1]

        for part in parts[:-1]:
            obj = obj.setdefault(part, {})

        obj.setdefault(parts[-1], set()).add(desc.cls)

    return result


def _generate_plugin_tree_overview(
    plugin_tree: dict | set[type[Plugin]],
    print_docs: bool,
    indent: int = 0,
) -> list[str]:
    """Create plugin overview with identations."""

    if isinstance(plugin_tree, set):
        return [
            textwrap.indent(
                get_plugin_overview(plugin, print_docs, print_docs),
                prefix=" " * indent,
            )
            for plugin in sorted(plugin_tree, key=lambda x: x.__module__)
        ]

    result = []
    for key in sorted(plugin_tree.keys()):
        result.append(textwrap.indent(key + ":", prefix=" " * indent) if key != "" else "OS plugins")
        result.extend(
            _generate_plugin_tree_overview(
                plugin_tree[key],
                print_docs,
                indent=indent + 2,
            )
        )

    return result


class PluginListPlugin(Plugin):
    """Plugin list plugin (so meta)."""

    def check_compatible(self) -> None:
        pass

    @export(output="none", cache=False)
    @arg("--docs", dest="print_docs", action="store_true", help="output docstrings")
    @arg("-j", "--json", action="store_true", help="output in JSON format")
    def plugins(self, print_docs: bool = False, json: bool = False) -> None:
        """Print all available plugins."""
        if json:
            print(generate_functions_json(), end="")
        else:
            print(generate_functions_overview(include_docs=print_docs))



from __future__ import annotations

import re
from datetime import datetime, timezone
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

re_field = re.compile(r"(.+)=(.+)")

COMMON_ELEMENTS = [
    ("datetime", "start_time"),
    ("datetime", "stop_time"),
    ("datetime", "created"),
    ("datetime", "modified"),
    ("datetime", "access"),
    ("path", "path"),
    ("path", "filename"),
    ("path", "create"),
    ("varint", "size"),
    ("string", "magic"),
    ("string", "size_of_image"),
    ("varint", "pe_checksum"),
    ("datetime", "link_date"),
    ("string", "linker_version"),
    ("string", "bin_file_version"),
    ("string", "bin_product_version"),
    ("string", "binary_type"),
    ("digest", "digest"),
    ("wstring", "file_version"),
    ("wstring", "company_name"),
    ("wstring", "file_description"),
    ("wstring", "legal"),
    ("string", "original_filename"),
    ("wstring", "product_name"),
    ("string", "product_version"),
    ("string", "pe_image"),
    ("string", "pe_subsystem"),
    ("string", "crc_checksum"),
    ("filesize", "filesize"),
    ("wstring", "longname"),
    ("string", "msi"),
]

AmcacheFileCreateRecord = TargetRecordDescriptor(
    "filesystem/windows/amcache/install/file_create",
    COMMON_ELEMENTS,
)

AmcacheArpCreateRecord = TargetRecordDescriptor(
    "filesystem/windows/amcache/install/arp_create",
    COMMON_ELEMENTS,
)


def _to_log_timestamp(timestamp: str) -> datetime:
    return datetime.strptime(timestamp, "%m/%d/%Y %H:%M:%S").replace(tzinfo=timezone.utc)


def create_record(
    description: AmcacheFileCreateRecord | AmcacheArpCreateRecord,
    filename: str,
    install_properties: dict[str, str],
    create: str,
    target: Target,
) -> TargetRecordDescriptor:
    return description(
        start_time=_to_log_timestamp(install_properties.get("starttime")),
        stop_time=_to_log_timestamp(install_properties.get("stoptime")),
        created=_to_log_timestamp(install_properties.get("created")),
        modified=_to_log_timestamp(install_properties.get("modified")),
        access=_to_log_timestamp(install_properties.get("lastaccessed")),
        link_date=_to_log_timestamp(install_properties.get("linkdate")),
        path=target.fs.path(install_properties.get("path")),
        filename=target.fs.path(filename),
        create=target.fs.path(create),
        size_of_image=install_properties.get("sizeofimage"),
        file_description=install_properties.get("filedescription"),
        size=int(install_properties.get("size", "0x00"), 16),
        digest=(
            None,
            install_properties.get("id")[4:],
            None,
        ),  # remove leading zeros from the entry to create a sha1 hash
        company_name=install_properties.get("companyname"),
        binary_type=install_properties.get("binarytype"),
        bin_product_version=install_properties.get("binproductversion"),
        bin_file_version=install_properties.get("binfileversion"),
        filesize=int(install_properties.get("filesize", "0"), 16),
        pe_image=install_properties.get("peimagetype"),
        product_version=install_properties.get("productversion"),
        crc_checksum=install_properties.get("crcchecksum"),
        legal=install_properties.get("legalcopyright"),
        magic=install_properties.get("magic"),
        linker_version=install_properties.get("linkerversion"),
        product_name=install_properties.get("productname"),
        pe_subsystem=install_properties.get("pesubsystem"),
        longname=install_properties.get("longname"),
        pe_checksum=int(install_properties.get("pechecksum", "0x00"), 16),
        _target=target,
    )


class AmcacheInstallPlugin(Plugin):
    """Amcache install log plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self.logs = self.target.resolve("%windir%/appcompat/programs/install")

    def check_compatible(self) -> None:
        if not self.logs.exists():
            raise UnsupportedPluginError("No amcache install logs found")

    @export(record=[AmcacheArpCreateRecord, AmcacheFileCreateRecord])
    def amcache_install(self) -> Iterator[AmcacheArpCreateRecord, AmcacheFileCreateRecord]:
        """Return the contents of the Amcache install log.

        The log file contains the changes an installer performed on the system.
        These only get created when the executable is an installer.
        """
        for f in self.logs.iterdir():
            install_properties = {}
            arp_created = []
            file_created = []
            for line in f.read_text("utf-16-le").splitlines():
                match = re_field.match(line.rstrip())
                if not match:
                    continue

                if match.group(1) == "FileCreate":
                    file_created.append(match.group(2))

                elif match.group(1) == "ArpCreate":
                    arp_created.append(match.group(2))

                else:
                    install_properties[match.group(1).lower()] = match.group(2)

            filename = str(f)
            for arp_create in arp_created:
                yield create_record(
                    AmcacheArpCreateRecord,
                    filename=filename,
                    install_properties=install_properties,
                    create=arp_create,
                    target=self.target,
                )
            for file_create in file_created:
                yield create_record(
                    AmcacheFileCreateRecord,
                    filename=filename,
                    install_properties=install_properties,
                    create=file_create,
                    target=self.target,
                )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.plugins.os.windows import ual
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


def test_ual_plugin(target_win: Target, fs_win: VirtualFilesystem) -> None:
    ual_dir = absolute_path("_data/plugins/os/windows/ual")

    fs_win.map_dir("Windows/System32/LogFiles/Sum", ual_dir)

    target_win.add_plugin(ual.UalPlugin)

    client_access_records = list(target_win.ual.client_access())
    assert len(client_access_records) == 106

    system_identity_records = list(target_win.ual.system_identities())
    assert len(system_identity_records) == 2

    role_access_records = list(target_win.ual.role_access())
    assert len(role_access_records) == 3

    virtual_machines_records = list(target_win.ual.virtual_machines())
    assert len(virtual_machines_records) == 0

    domains_seen_records = list(target_win.ual.domains_seen())
    assert len(domains_seen_records) == 12

    ual_all_records = list(target_win.ual())
    assert len(ual_all_records) == 123



from __future__ import annotations

import textwrap
from datetime import datetime, timedelta, timezone
from io import BytesIO
from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.apps.webserver.apache import (
    LOG_FORMAT_ACCESS_COMBINED,
    LOG_FORMAT_ACCESS_COMMON,
    LOG_FORMAT_ACCESS_VHOST_COMBINED,
    ApachePlugin,
    LogFormat,
)
from tests._utils import absolute_path

if TYPE_CHECKING:
    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.mark.parametrize(
    ("log_line", "expected_log_format"),
    [
        pytest.param(
            (
                '127.0.0.1 - - [19/Dec/2022:17:25:12 +0100] "GET / HTTP/1.1" 304 247 "-" "Mozilla/5.0 '
                "(Windows NT 10.0; Win64; x64); AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 "
                'Safari/537.36"'
            ),
            LOG_FORMAT_ACCESS_COMBINED,
            id="combined",
        ),
        pytest.param(
            (
                '2001:0db8:85a3:0000:0000:8a2e:0370:7334 - - [20/Dec/2022:15:18:01 +0100] "GET / HTTP/1.1" 200 '
                '1126 "-" "curl/7.81.0"'
            ),
            LOG_FORMAT_ACCESS_COMBINED,
            id="combined-ipv6",
        ),
        pytest.param(
            (
                'example.com:80 127.0.0.1 - - [19/Dec/2022:17:25:40 +0100] "GET / HTTP/1.1" 200 312 '
                '"-" "Mozilla/5.0 (Windows NT 10.0; Win64; x64); AppleWebKit/537.36 (KHTML, like Gecko) '
                'Chrome/108.0.0.0 Safari/537.36"'
            ),
            LOG_FORMAT_ACCESS_VHOST_COMBINED,
            id="vhost-combined",
        ),
        pytest.param(
            '127.0.0.1 - - [19/Dec/2022:17:25:40 +0100] "GET / HTTP/1.1" 200 312',
            LOG_FORMAT_ACCESS_COMMON,
            id="common",
        ),
    ],
)
def test_infer_access_log_format(log_line: str, expected_log_format: LogFormat) -> None:
    """test if we infer the access log format for given log lines correctly."""
    assert ApachePlugin.infer_access_log_format(log_line) == expected_log_format


def test_txt(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/apache/access.log")
    fs_unix.map_file("var/log/apache2/access.log", data_file)

    target_unix.add_plugin(ApachePlugin)
    results = list(target_unix.apache.access())

    assert len(results) == 6


def test_gz(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/apache/access.log.gz")
    fs_unix.map_file("var/log/apache2/access.log.1.gz", data_file)

    target_unix.add_plugin(ApachePlugin)
    results = list(target_unix.apache.access())

    assert len(results) == 4


def test_access_bz2(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/apache/access.log.bz2")
    fs_unix.map_file("var/log/apache2/access.log.1.bz2", data_file)

    target_unix.add_plugin(ApachePlugin)
    results = list(target_unix.apache.access())

    assert len(results) == 4


def test_xampp_unix(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/apache/access.log")
    fs_unix.map_file("opt/lampp/logs/access.log", data_file)

    target_unix.add_plugin(ApachePlugin)
    results = list(target_unix.apache.access())

    assert len(results) == 6


def test_xampp_windows(target_win: Target, fs_win: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/apache/access.log")
    fs_win.map_file("xampp/apache/logs/access.log", data_file)

    target_win.add_plugin(ApachePlugin)
    results = list(target_win.apache.access())

    assert len(results) == 6


def test_all_access_log_formats(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    tz = timezone(timedelta(hours=1))
    data_file = absolute_path("_data/plugins/apps/webserver/apache/access.log")
    fs_unix.map_file("var/log/apache2/access.log", data_file)

    target_unix.add_plugin(ApachePlugin)
    results = list(target_unix.apache.access())

    assert len(results) == 6

    combined_log = results[0]
    assert combined_log.ts == datetime(2022, 12, 19, 17, 6, 19, tzinfo=tz)
    assert combined_log.status_code == 200
    assert combined_log.remote_ip == "1.2.3.4"
    assert combined_log.remote_user is None
    assert combined_log.method == "GET"
    assert combined_log.uri == "/"
    assert combined_log.protocol == "HTTP/1.1"
    assert combined_log.referer == "Sample referer"
    assert (
        combined_log.useragent
        == "Mozilla/5.0 (Windows NT 10.0; Win64; x64); AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0 "
        "Safari/537.36"
    )

    vhost_combined_log = results[1]
    assert vhost_combined_log.ts == datetime(2022, 12, 19, 17, 25, 40, tzinfo=tz)
    assert vhost_combined_log.status_code == 200
    assert vhost_combined_log.remote_ip == "1.2.3.4"
    assert vhost_combined_log.remote_user is None
    assert vhost_combined_log.method == "GET"
    assert vhost_combined_log.uri == "/index.html"
    assert vhost_combined_log.protocol == "HTTP/1.1"
    assert vhost_combined_log.referer is None
    assert (
        vhost_combined_log.useragent
        == "Mozilla/5.0 (Windows NT 10.0; Win64; x64); AppleWebKit/537.36 (KHTML, like Gecko) Chrome/107.0.0.0 "
        "Safari/537.36"
    )

    common_log = results[2]
    assert common_log.ts == datetime(2022, 12, 19, 17, 25, 48, tzinfo=tz)
    assert common_log.status_code == 200
    assert common_log.remote_ip == "4.3.2.1"
    assert common_log.remote_user is None
    assert common_log.method == "GET"
    assert common_log.uri == "/"
    assert common_log.protocol == "HTTP/1.1"
    assert common_log.referer is None
    assert common_log.useragent is None

    ipv6_log = results[3]
    assert ipv6_log.remote_ip == "2001:0db8:85a3:0000:0000:8a2e:0370:7334"
    assert ipv6_log.protocol is None
    assert ipv6_log.bytes_sent == 0


def test_logrotate(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/apache/access.log")
    fs_unix.map_file("var/log/apache2/access.log", data_file)
    fs_unix.map_file("var/log/apache2/access.log.1", data_file)
    fs_unix.map_file("var/log/apache2/access.log.2", data_file)
    fs_unix.map_file("var/log/apache2/access.log.3", data_file)

    target_unix.add_plugin(ApachePlugin)

    assert len(target_unix.apache.access_paths) == 4
    assert len(target_unix.apache.error_paths) == 0


def test_custom_config(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    fs_unix.map_file_fh(
        "/etc/apache2/apache2.conf", BytesIO(b'CustomLog "/very/custom/log/location/access.log" common')
    )
    fs_unix.map_file_fh("/very/custom/log/location/access.log", BytesIO(b"Foo"))
    fs_unix.map_file_fh("/very/custom/log/location/access.log.1", BytesIO(b"Foo1"))
    fs_unix.map_file_fh("/very/custom/log/location/access.log.2", BytesIO(b"Foo2"))
    fs_unix.map_file_fh("/very/custom/log/location/access.log.3", BytesIO(b"Foo3"))

    target_unix.add_plugin(ApachePlugin)

    assert sorted(map(str, target_unix.apache.access_paths)) == [
        "/very/custom/log/location/access.log",
        "/very/custom/log/location/access.log.1",
        "/very/custom/log/location/access.log.2",
        "/very/custom/log/location/access.log.3",
    ]
    assert len(target_unix.apache.error_paths) == 0


def test_config_commented_logs(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    config = """
    # CustomLog "/custom/log/location/old.log" common
    CustomLog "/custom/log/location/new.log" common
    # ErrorLog "/custom/log/location//old_error.log"
    ErrorLog "/custom/log/location//new_error.log"
    """

    fs_unix.map_file_fh("etc/httpd/conf/httpd.conf", BytesIO(textwrap.dedent(config).encode()))
    fs_unix.map_file_fh("custom/log/location/new.log", BytesIO(b"New"))
    fs_unix.map_file_fh("custom/log/location/old.log", BytesIO(b"Old"))
    fs_unix.map_file_fh("custom/log/location/old_error.log", BytesIO(b"Old"))
    fs_unix.map_file_fh("custom/log/location/new_error.log", BytesIO(b"New"))

    target_unix.add_plugin(ApachePlugin)

    assert sorted(map(str, target_unix.apache.access_paths)) == [
        "/custom/log/location/new.log",
        "/custom/log/location/old.log",
    ]

    assert sorted(map(str, target_unix.apache.error_paths)) == [
        "/custom/log/location/new_error.log",
        "/custom/log/location/old_error.log",
    ]


def test_config_vhosts_httpd(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we detect httpd CustomLog and ErrorLog directives using IncludeOptional configuration."""
    config = """
    ServerRoot "/etc/httpd"
    IncludeOptional conf/vhosts/*/*.conf
    """

    # '/etc/httpd/conf/vhosts/*/*.conf'
    vhost_config_1 = """
    CustomLog "/custom/log/location/vhost_1.log" common
    ErrorLog "/custom/log/location/vhost_1.error"
    """
    vhost_config_2 = """
    CustomLog "/custom/log/location/vhost_2.log" combined
    ErrorLog "/custom/log/location/vhost_2.error"
    """
    fs_unix.map_file_fh("etc/httpd/conf/httpd.conf", BytesIO(textwrap.dedent(config).encode()))
    fs_unix.map_file_fh("etc/httpd/conf/vhosts/host1/host1.conf", BytesIO(textwrap.dedent(vhost_config_1).encode()))
    fs_unix.map_file_fh("etc/httpd/conf/vhosts/host2/host2.conf", BytesIO(textwrap.dedent(vhost_config_2).encode()))
    fs_unix.map_file_fh("custom/log/location/vhost_1.log", BytesIO(b"Log 1"))
    fs_unix.map_file_fh("custom/log/location/vhost_2.log", BytesIO(b"Log 2"))
    fs_unix.map_file_fh("custom/log/location/vhost_1.error", BytesIO(b"Err 1"))
    fs_unix.map_file_fh("custom/log/location/vhost_2.error", BytesIO(b"Err 2"))

    target_unix.add_plugin(ApachePlugin)

    assert sorted(map(str, target_unix.apache.access_paths)) == [
        "/custom/log/location/vhost_1.log",
        "/custom/log/location/vhost_2.log",
    ]
    assert sorted(map(str, target_unix.apache.error_paths)) == [
        "/custom/log/location/vhost_1.error",
        "/custom/log/location/vhost_2.error",
    ]


def test_config_vhosts_apache2(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we detect apache2 CustomLog and ErrorLog directives using IncludeOptional configuration."""
    config = r"""
    ServerRoot "/etc/apache2"
    ErrorLog ${APACHE_LOG_DIR}/error.log
    Include example.conf
    IncludeOptional conf-enabled/*.conf
    IncludeOptional sites-enabled/*.conf
    """
    fs_unix.map_file_fh("/etc/apache2/apache2.conf", BytesIO(textwrap.dedent(config).encode()))
    fs_unix.map_file_fh("/path/to/apache/logs/error.log", BytesIO())

    envvars = r"""
    export APACHE_LOG_DIR="/path/to/apache/logs"
    """
    fs_unix.map_file_fh("/etc/apache2/envvars", BytesIO(textwrap.dedent(envvars).encode()))

    enabled_conf = r"""
    CustomLog ${APACHE_LOG_DIR}/example_access.log custom
    """
    fs_unix.map_file_fh("/etc/apache2/conf-enabled/example.conf", BytesIO(textwrap.dedent(enabled_conf).encode()))
    fs_unix.map_file_fh("/path/to/apache/logs/example_access.log.1.gz", BytesIO())

    site_conf = """
    <VirtualHost *:80>
        ServerName example.com
        DocumentRoot /var/www/html
        errorlog /path/to/virtualhost/log/error.log
        customlog /path/to/virtualhost/log/access.log custom
    </VirtualHost>
    """
    fs_unix.map_file_fh("/etc/apache2/sites-enabled/example.conf", BytesIO(textwrap.dedent(site_conf).encode()))
    fs_unix.map_file_fh("/path/to/virtualhost/log/error.log.1", BytesIO())
    fs_unix.map_file_fh("/path/to/virtualhost/log/access.log.1", BytesIO())

    disabled_conf = """
    CustomLog /path/to/disabled/access.log custom
    """
    fs_unix.map_file_fh(
        "/etc/apache2/sites-available/disabled-site.conf", BytesIO(textwrap.dedent(disabled_conf).encode())
    )
    fs_unix.map_file_fh("/path/to/disabled/access.log.2", BytesIO())

    fs_unix.map_file_fh("/var/log/apache2/some-other-vhost-old-log.access.log", BytesIO())
    fs_unix.map_file_fh("/var/log/apache2/access.log", BytesIO())
    fs_unix.map_file_fh("/var/log/apache2/error.log", BytesIO())

    target_unix.add_plugin(ApachePlugin)

    assert sorted(map(str, target_unix.apache.access_paths)) == [
        "/path/to/apache/logs/example_access.log.1.gz",
        "/path/to/disabled/access.log.2",
        "/path/to/virtualhost/log/access.log.1",
        "/var/log/apache2/access.log",
        "/var/log/apache2/some-other-vhost-old-log.access.log",
    ]

    assert sorted(map(str, target_unix.apache.error_paths)) == [
        "/path/to/apache/logs/error.log",
        "/path/to/virtualhost/log/error.log.1",
        "/var/log/apache2/error.log",
    ]


def test_error_txt(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/apache/error.log")
    fs_unix.map_file("var/log/apache2/error.log", data_file)

    target_unix.add_plugin(ApachePlugin)
    results = list(target_unix.apache.error())

    assert len(results) == 3
    assert all(str(record.source) == "/var/log/apache2/error.log" for record in results)


def test_all_error_log_formats(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    data_file = absolute_path("_data/plugins/apps/webserver/apache/error.log")
    fs_unix.map_file("var/log/apache2/error.log", data_file)

    target_unix.add_plugin(ApachePlugin)
    results = list(target_unix.apache.error())

    assert len(results) == 3

    no_client_log = results[0]
    assert no_client_log.ts == datetime(2022, 12, 13, 13, 7, 21, 131489, tzinfo=timezone.utc)
    assert no_client_log.error_code == "AH00163"
    assert no_client_log.module == "mpm_prefork"
    assert no_client_log.level == "notice"
    assert no_client_log.pid == 218
    assert no_client_log.message == (
        "Apache/2.4.54 (Unix) mod_perl/2.0.10 Perl/v5.30.2 configured -- resuming normal operations"
    )

    client_log = results[1]
    assert client_log.ts == datetime(2011, 5, 12, 8, 28, 57, 652118, tzinfo=timezone.utc)
    assert client_log.error_code is None
    assert client_log.remote_ip == "::1"
    assert client_log.module == "core"
    assert client_log.level == "error"
    assert client_log.pid == 8777
    assert client_log.message == "File does not exist: /usr/local/apache2/htdocs/favicon.ico"

    log_with_source = results[2]
    assert log_with_source.ts == datetime(2019, 2, 1, 22, 3, 8, 320285, tzinfo=timezone.utc)
    assert log_with_source.error_code == "AH01626"
    assert log_with_source.remote_ip == "172.17.0.1"
    assert log_with_source.module == "authz_core"
    assert log_with_source.level == "debug"
    assert log_with_source.pid == 7
    assert log_with_source.message == "authorization result of <RequireAny>: granted"
    assert log_with_source.error_source == "mod_authz_core.c(820)"


def test_apache_virtual_hosts(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we can find and parse virtual host configurations correctly."""

    fs_unix.map_file_fh("/etc/apache2/apache2.conf", BytesIO(b'ServerRoot "/etc/apache2"\n'))

    site = r"""
    <VirtualHost *:443>
        ServerName example.com
        DocumentRoot /var/www/html
        ErrorLog ${APACHE_LOG_DIR}/error.log
        CustomLog ${APACHE_LOG_DIR}/access.log combined
    </VirtualHost>
    <Virtualhost 127.0.0.1:80>
        documentroot /path/to/other/html
    </Virtualhost>
    """
    fs_unix.map_file_fh("/etc/apache2/sites-available/example.conf", BytesIO(textwrap.dedent(site).encode()))

    target_unix.add_plugin(ApachePlugin)

    records = list(target_unix.apache.hosts())
    assert len(records) == 2

    assert records[0].ts
    assert records[0].server_name == "example.com"
    assert records[0].server_port == 443
    assert records[0].root_path == "/var/www/html"
    assert records[0].access_log_config == r"${APACHE_LOG_DIR}/access.log"
    assert records[0].error_log_config == r"${APACHE_LOG_DIR}/error.log"
    assert records[0].source == "/etc/apache2/sites-available/example.conf"

    assert records[1].ts
    assert records[1].server_name == "127.0.0.1"
    assert records[1].server_port == 80
    assert records[1].root_path == "/path/to/other/html"
    assert records[1].source == "/etc/apache2/sites-available/example.conf"


def test_apache_access_format_malformed_regression(target_unix: Target, fs_unix: VirtualFilesystem) -> None:
    """Test if we correctly detect and parse some "malformed" access log formats."""

    # Combined format without e.g. 'GET / HTTP/1.1' but instead '"-"'.
    combined_without_http_phrase = b'1.2.3.4 - - [11/Nov/2025:12:34:56 +0100] "-" 418 - "-" "-"'

    # Common format without a response size in bytes e.g. 1234 but instead '-'.
    common_without_response_size = b'5.6.7.8 - - [31/Dec/2025:01:23:45 +0100] "GET / HTTP/1.1" 418 -'

    fs_unix.map_file_fh(
        "/var/log/apache2/access.log", BytesIO(combined_without_http_phrase + b"\n" + common_without_response_size)
    )

    target_unix.add_plugin(ApachePlugin)
    results = list(target_unix.apache.access())

    assert len(results) == 2

    assert results[0].ts == datetime(2025, 11, 11, 11, 34, 56, 0, tzinfo=timezone.utc)
    assert results[0].remote_user is None
    assert results[0].remote_ip == "1.2.3.4"
    assert results[0].local_ip is None
    assert results[0].pid is None
    assert results[0].method is None
    assert results[0].uri is None
    assert results[0].protocol is None
    assert results[0].status_code == 418
    assert results[0].bytes_sent == 0
    assert results[0].referer is None
    assert results[0].useragent is None
    assert results[0].response_time_ms is None
    assert results[0].source == "/var/log/apache2/access.log"

    assert results[1].ts == datetime(2025, 12, 31, 0, 23, 45, 0, tzinfo=timezone.utc)
    assert results[1].remote_user is None
    assert results[1].remote_ip == "5.6.7.8"
    assert results[1].local_ip is None
    assert results[1].pid is None
    assert results[1].method == "GET"
    assert results[1].uri == "/"
    assert results[1].protocol == "HTTP/1.1"
    assert results[1].status_code == 418
    assert results[1].bytes_sent == 0
    assert results[1].referer is None
    assert results[1].useragent is None
    assert results[1].response_time_ms is None
    assert results[1].source == "/var/log/apache2/access.log"



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.loader import Loader
from dissect.target.loaders.dir import find_and_map_dirs, find_dirs
from dissect.target.plugin import OperatingSystem

if TYPE_CHECKING:
    from pathlib import Path

    from dissect.target.target import Target


class TaniumLoader(Loader):
    """Load Tanium forensic image format files."""

    @staticmethod
    def detect(path: Path) -> bool:
        # A Tanium package is very similar to a Kape package.
        # The 'file' folder contains the disk data acquired using the Tanium client.
        # The 'collector' folder contains parsed, interpreted and collected host info
        # TIMESTAMP-HOSTNAME/
        #   collector/
        #       shell_bags.txt
        #       network_connections.txt
        #       processes.txt
        #       handles.txt
        #   file/
        #       C/
        #           Windows/
        #           Users/
        #           ...../
        #       D/
        file_path = path.joinpath("file")
        os_type, dirs = find_dirs(file_path)
        if file_path and os_type == OperatingSystem.WINDOWS:
            for path in dirs:
                # Tanium doesn't have the correct filenames for several files, like $J
                if path.joinpath("$Extend/$UsnJrnl_$J").exists():
                    return True

        return False

    def map(self, target: Target) -> None:
        find_and_map_dirs(
            target,
            self.absolute_path.joinpath("file"),
            sds_path="$Secure_$SDS",
            usnjrnl_path="$Extend/$UsnJrnl_$J",
        )



from __future__ import annotations

import hashlib
from typing import TYPE_CHECKING, BinaryIO

if TYPE_CHECKING:
    from hashlib._hashlib import HASH

BUFFER_SIZE = 32768


def _hash(fh: BinaryIO, ctx: HASH | list[HASH]) -> tuple[str]:
    if not isinstance(ctx, list):
        ctx = [ctx]

    ctx = [c() for c in ctx]
    data = fh.read(BUFFER_SIZE)
    while data:
        [c.update(data) for c in ctx]
        data = fh.read(BUFFER_SIZE)

    return tuple(c.hexdigest() for c in ctx)


def md5(fh: BinaryIO) -> tuple[str]:
    return _hash(fh, hashlib.md5)[0]


def sha1(fh: BinaryIO) -> tuple[str]:
    return _hash(fh, hashlib.sha1)[0]


def sha256(fh: BinaryIO) -> tuple[str]:
    return _hash(fh, hashlib.sha256)[0]


def common(fh: BinaryIO) -> tuple[str]:
    return _hash(fh, [hashlib.md5, hashlib.sha1, hashlib.sha256])


def custom(fh: BinaryIO, algos: list[str | HASH]) -> tuple[str]:
    ctx = [getattr(hashlib, h) for h in algos] if isinstance(algos[0], str) else algos

    return _hash(fh, ctx)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.cstruct import cstruct
from dissect.util.ts import wintimestamp

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import create_extended_descriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target

RecycleBinRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/filesystem/recyclebin",
    [
        ("datetime", "ts"),
        ("path", "path"),
        ("filesize", "filesize"),
        ("path", "deleted_path"),
        ("string", "source"),
    ],
)

recyclebin_def = """
struct header_v1 {
    int64    version;
    int64    file_size;
    int64    timestamp;
    wchar    filename[260];
};
struct header_v2 {
    int64    version;
    int64    file_size;
    int64    timestamp;
    int32    filename_length;
    wchar    filename[filename_length];
};
"""

c_recyclebin = cstruct().load(recyclebin_def)


class RecyclebinPlugin(Plugin):
    """Recyclebin plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self.recyclebin_paths = []
        for fs_entry in self.target.fs.path("/").iterdir():
            if self._is_valid_recyclebin(fs_entry):
                self.recyclebin_paths.append(fs_entry.joinpath("$recycle.bin"))

    def check_compatible(self) -> None:
        if not self.recyclebin_paths:
            raise UnsupportedPluginError("No recycle bins found")

    def _is_valid_recyclebin(self, path: TargetPath) -> bool:
        """Checks wether it is a valid recycle bin path.

        The sysvol is skipped so that there are no duplicated with drive letter that maps to sysvol.
        """

        return path.name != "sysvol" and path.joinpath("$recycle.bin").exists()

    @export(record=RecycleBinRecord)
    def recyclebin(self) -> Iterator[RecycleBinRecord]:
        """
        Return files located in the recycle bin ($Recycle.Bin).

        Yields RecycleBinRecords with fields:

        .. code-block:: text

          hostname (string): The target hostname
          domain (string): The target domain
          ts (datetime): The time of deletion
          path (uri): The file original location before deletion
          filesize (filesize): Filesize of the deleted file
          sid (string): SID of the user deleted the file, parsed from $I filepath
          user (string): Username matching SID, lookup using Dissect user plugin
          deleted_path (uri): Location of the deleted file after deletion $R file
          source (uri): Location of $I meta file on disk
        """

        for recyclebin in self.recyclebin_paths:
            yield from self.read_recycle_bin(recyclebin)

    def _is_recycle_file(self, path: TargetPath) -> bool:
        """Check wether the path is a recycle bin metadata file."""
        return path.name and path.name.lower().startswith("$i")

    def read_recycle_bin(self, bin_path: TargetPath) -> Iterator[RecycleBinRecord]:
        if self._is_recycle_file(bin_path):
            yield self.read_bin_file(bin_path)
            return

        if bin_path.is_dir():
            for new_file in bin_path.iterdir():
                yield from self.read_recycle_bin(new_file)

    def read_bin_file(self, bin_path: TargetPath) -> RecycleBinRecord:
        data = bin_path.read_bytes()

        header = self.select_header(data)
        entry = header(data)

        sid = self.find_sid(bin_path)
        source_path = str(bin_path).lstrip("/")
        deleted_path = str(bin_path.parent / bin_path.name.replace("/$i", "/$r")).lstrip("/")

        user_details = self.target.user_details.find(sid=sid)
        user = user_details.user if user_details else None

        return RecycleBinRecord(
            ts=wintimestamp(entry.timestamp),
            path=self.target.fs.path(entry.filename.rstrip("\x00")),
            source=self.target.fs.path(source_path),
            filesize=entry.file_size,
            deleted_path=self.target.fs.path(deleted_path),
            _target=self.target,
            _user=user,
        )

    def find_sid(self, path: TargetPath) -> str:
        parent_path = path.parent
        if parent_path.name.lower() == "$recycle.bin":
            return "unknown"
        return parent_path.name

    def select_header(self, data: bytes) -> c_recyclebin.header_v1 | c_recyclebin.header_v2:
        """Selects the correct header based on the version field in the header."""

        header_version = c_recyclebin.uint64(data[:8])
        if header_version == 2:
            return c_recyclebin.header_v2
        return c_recyclebin.header_v1



from __future__ import annotations

from pathlib import Path
from tempfile import TemporaryDirectory

from dissect.target.helpers import config


def test_load_config() -> None:
    # FS layout:
    #
    # temp_dir1
    #   config_file
    #   symlink_dir2 -> ../temp_dir2
    # temp_dir2

    with TemporaryDirectory() as temp_dir1, TemporaryDirectory() as temp_dir2:
        # create symlink in temp_dir1 pointing to temp_dir2
        symlink = Path(temp_dir1).joinpath("symlink")
        symlink.symlink_to(temp_dir2)

        config_file = Path(temp_dir1).joinpath(config.CONFIG_NAME)
        config_file.write_text('CONFIG_FILE = "found"')

        result = config.load(symlink)
        assert result.CONFIG_FILE == "found"



from __future__ import annotations

import stat
from io import BytesIO
from typing import TYPE_CHECKING, BinaryIO

from dissect.sql import sqlite3
from dissect.util.stream import BufferedStream

from dissect.target.filesystem import (
    Filesystem,
    VirtualDirectory,
    VirtualFile,
    VirtualFilesystem,
)
from dissect.target.helpers import fsutil
from dissect.target.plugins.os.unix._os import OperatingSystem, export
from dissect.target.plugins.os.unix.linux.debian._os import DebianPlugin

if TYPE_CHECKING:
    from typing_extensions import Self

    from dissect.target.target import Target


class ProxmoxPlugin(DebianPlugin):
    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        for fs in target.filesystems:
            if fs.exists("/etc/pve") or fs.exists("/var/lib/pve"):
                return fs

        return None

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> Self:
        obj = super().create(target, sysvol)

        if (config_db := target.fs.path("/var/lib/pve-cluster/config.db")).exists():
            with config_db.open("rb") as fh:
                vfs = _create_pmxcfs(fh, obj.hostname)

            target.fs.mount("/etc/pve", vfs)

        return obj

    @export(property=True)
    def version(self) -> str:
        """Returns Proxmox VE version with underlying OS release."""

        for pkg in self.target.dpkg.status():
            if pkg.name == "proxmox-ve":
                distro_name = self._os_release.get("PRETTY_NAME", "")
                return f"{pkg.name} {pkg.version} ({distro_name})"
        return None

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.PROXMOX.value


DT_DIR = 4
DT_REG = 8


def _create_pmxcfs(fh: BinaryIO, hostname: str | None = None) -> VirtualFilesystem:
    # https://pve.proxmox.com/wiki/Proxmox_Cluster_File_System_(pmxcfs)
    db = sqlite3.SQLite3(fh)

    entries = {row.inode: row for row in db.table("tree")}

    vfs = VirtualFilesystem()
    for entry in entries.values():
        if entry.type == DT_DIR:
            cls = ProxmoxConfigDirectoryEntry
        elif entry.type == DT_REG:
            cls = ProxmoxConfigFileEntry
        else:
            raise ValueError(f"Unknown pmxcfs file type: {entry.type}")

        parts = []
        current = entry
        while current.parent != 0:
            parts.append(current.name)
            current = entries[current.parent]
        parts.append(current.name)

        path = "/".join(parts[::-1])
        vfs.map_file_entry(path, cls(vfs, path, entry))

    if hostname:
        node_root = vfs.path(f"nodes/{hostname}")
        vfs.symlink(str(node_root), "local")
        vfs.symlink(str(node_root / "lxc"), "lxc")
        vfs.symlink(str(node_root / "openvz"), "openvz")
        vfs.symlink(str(node_root / "qemu-server"), "qemu-server")

    # TODO: .version, .members, .vmlist, maybe .clusterlog and .rrd?

    return vfs


class ProxmoxConfigFileEntry(VirtualFile):
    def open(self) -> BinaryIO:
        return BufferedStream(BytesIO(self.entry.data or b""))

    def lstat(self) -> fsutil.stat_result:
        # ['mode', 'addr', 'dev', 'nlink', 'uid', 'gid', 'size', 'atime', 'mtime', 'ctime']
        return fsutil.stat_result(
            [
                stat.S_IFREG | 0o640,
                self.entry.inode,
                id(self.fs),
                1,
                0,
                0,
                len(self.entry.data) if self.entry.data else 0,
                0,
                self.entry.mtime,
                0,
            ]
        )


class ProxmoxConfigDirectoryEntry(VirtualDirectory):
    def __init__(self, fs: VirtualFilesystem, path: str, entry: sqlite3.Row):
        super().__init__(fs, path)
        self.entry = entry

    def lstat(self) -> fsutil.stat_result:
        """Return the stat information of the given path, without resolving links."""
        # ['mode', 'addr', 'dev', 'nlink', 'uid', 'gid', 'size', 'atime', 'mtime', 'ctime']
        return fsutil.stat_result(
            [
                stat.S_IFDIR | 0o755,
                self.entry.inode,
                id(self.fs),
                1,
                0,
                0,
                0,
                0,
                self.entry.mtime,
                0,
            ]
        )



from __future__ import annotations

import platform
from typing import TYPE_CHECKING, BinaryIO

from dissect.cstruct import cstruct
from dissect.util.stream import RangeStream

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import export
from dissect.target.plugins.os.unix.locate.locate import BaseLocatePlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

try:
    from zstandard import (
        DECOMPRESSION_RECOMMENDED_OUTPUT_SIZE,
        ZstdCompressionDict,
        ZstdDecompressor,
    )

    HAS_ZSTD = True
except ImportError:
    HAS_ZSTD = False

# Resource: https://git.sesse.net/?p=plocate @ db.h
plocate_def = """
#define MAGIC 0x00706c6f63617465             /* b'/x00plocate' */

struct header {
    uint32_t version;
    uint32_t hashtable_size;
    uint32_t extra_ht_slots;
    uint32_t num_docids;
    uint64_t hash_table_offset_bytes;
    uint64_t filename_index_offset_bytes;

    /* Version 1 and up only. */
    uint32_t max_version;   // Nominally 1 or 2, but can be increased if more features are added in a backward-compatible way.
    uint32_t zstd_dictionary_length_bytes;
    uint64_t zstd_dictionary_offset_bytes;

    /* Only if max_version >= 2, and only relevant for updatedb. */
    uint64_t directory_data_length_bytes;
    uint64_t directory_data_offset_bytes;
    uint64_t next_zstd_dictionary_length_bytes;
    uint64_t next_zstd_dictionary_offset_bytes;
    uint64_t conf_block_length_bytes;
    uint64_t conf_block_offset_bytes;

    // Only if max_version >= 2.
    uint8_t check_visibility;
    char padding[7];                         /* padding for alignment */
};

struct file {
    char path[];
};
"""  # noqa : E501

PLocateRecord = TargetRecordDescriptor(
    "linux/locate/plocate",
    [
        ("path", "path"),
        ("path", "source"),
    ],
)

c_plocate = cstruct().load(plocate_def)


class PLocateFile:
    """Parser for plocate files.

    The ``plocate.db`` file contains a hashtable and trigrams to enable quick lookups of filenames.

    We've implemented a few methods to gather those for possible future use, but for the PLocatePlugin
    we're only interested in the filepaths stored in the database. Hence we don't use these methods.

    Roughly speaking, the plocate.db file has the following structure:
        - ``header`` (0x70 bytes)
        - zstd compressed ``filename``s (until start of ``filename_index_offset_bytes``),
          possibly including a dictionary
        - hashtables (offset and length in ``header``)
        - directory data (offset and length in ``header``)
        - possible zstd dictionary (offset and length in ``header``)
        - configuration block (offset and length in ``header``)

    No documentation other than the source code is available on the format of this file.

    References:
        - https://git.sesse.net/?p=plocate
    """

    HEADER_SIZE = 0x70  # 0x8 bytes magic + 0x68 bytes header
    NUM_OVERFLOW_SLOTS = 16
    TRIGRAM_SIZE_BYTES = 16
    DOCID_SIZE_BYTES = 8

    def __init__(self, fh: BinaryIO):
        self.fh = fh

        magic = int.from_bytes(self.fh.read(8), byteorder="big")
        if magic != c_plocate.MAGIC:
            raise ValueError(f"Invalid plocate file magic. Expected b'/x00plocate', got {magic}")

        self.header = c_plocate.header(self.fh)
        self.dict_data = None

        if self.header.zstd_dictionary_offset_bytes:
            self.dict_data = ZstdCompressionDict(self.fh.read(self.header.zstd_dictionary_length_bytes))

        self.compressed_length_bytes = (
            self.header.filename_index_offset_bytes - self.HEADER_SIZE - self.header.zstd_dictionary_length_bytes
        )
        self.ctx = ZstdDecompressor(dict_data=self.dict_data)
        self.buf = RangeStream(self.fh, self.fh.tell(), self.compressed_length_bytes)

    def __iter__(self) -> Iterator[PLocateFile]:
        # NOTE: This is a workaround for a PyPy bug
        # We don't know what breaks, but PyPy + zstandard = unhappy times
        # You just get random garbage data back instead of the decompressed data
        # This weird dance of using a decompressobj and unused data is the only way that seems to work
        # It's more expensive on memory, but at least it doesn't break
        if platform.python_implementation() == "PyPy":
            buf = self.buf.read()

            def reader(ctx: ZstdDecompressor) -> Iterator[bytes]:
                obj = ctx.decompressobj()

                yield obj.decompress(buf)
                while unused_data := obj.unused_data:
                    obj = self.ctx.decompressobj()
                    yield obj.decompress(unused_data)

            it = reader(self.ctx)
        else:
            # NOTE: The end of a zstandard frame does not include a final `0x00`.
            # This causes the c_plocate `file` struct to parse the last path and the first path on the next frame as one
            # since cstruct will read it across frame boundaries waiting for a `0x00`.
            def reader() -> Iterator[bytes]:
                with self.ctx.stream_reader(self.buf) as reader:
                    while chunk := reader.read(DECOMPRESSION_RECOMMENDED_OUTPUT_SIZE):
                        yield chunk

            it = reader()

        for chunk in it:
            for path in chunk.split(b"\x00"):
                yield path.decode(errors="surrogateescape")

    def filename_index(self) -> bytes:
        """Return the filename index of the plocate.db file."""
        self.fh.seek(self.header.filename_index_offset_bytes)
        num_docids = self.header.num_docids
        filename_index_size = num_docids * self.DOCID_SIZE_BYTES
        return self.fh.read(filename_index_size)

    def hashtable(self) -> bytes:
        """Return the hashtable of the plocate.db file."""
        self.fh.seek(self.header.hash_table_offset_bytes)
        hashtable_size = (self.header.hashtable_size + self.NUM_OVERFLOW_SLOTS + 1) * self.TRIGRAM_SIZE_BYTES
        return self.fh.read(hashtable_size)


class PLocatePlugin(BaseLocatePlugin):
    """Unix plocate plugin."""

    __namespace__ = "plocate"

    path = "/var/lib/plocate/plocate.db"

    def check_compatible(self) -> None:
        if not self.target.fs.path(self.path).exists():
            raise UnsupportedPluginError(f"No plocate.db file found at {self.path}")

        if not HAS_ZSTD:
            raise UnsupportedPluginError(
                "Please install `python-zstandard` or `pip install zstandard` to use the PLocatePlugin"
            )

    @export(record=PLocateRecord)
    def locate(self) -> Iterator[PLocateRecord]:
        """Yield file and directory names from the plocate.db.

        ``plocate`` is the default package on Ubuntu 22 and newer to locate files.
        It replaces ``mlocate`` and GNU ``locate``.

        References:
            - https://manpages.debian.org/testing/plocate/plocate.1.en.html
            - https://git.sesse.net/?p=plocate
        """
        plocate = self.target.fs.path(self.path)
        plocate_file = PLocateFile(plocate.open())

        for path in plocate_file:
            yield PLocateRecord(
                path=self.target.fs.path(path),
                source=self.path,
            )



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING
from unittest.mock import patch

from dissect.target.filesystem import VirtualFilesystem
from dissect.target.tools.mount import main as target_mount
from dissect.target.volume import Volume

if TYPE_CHECKING:
    import pytest

    from dissect.target.target import Target


def test_duplicate_volume_name(target_bare: Target, monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.setattr("sys.argv", ["target-mount", "mock-target", "mock-mount"])
        m.setattr("dissect.target.tools.mount.HAS_FUSE", True)

        with (
            patch("dissect.target.tools.mount.Target") as MockTarget,
            patch("dissect.target.tools.mount.FUSE", create=True) as MockFUSE,
            patch("dissect.target.tools.mount.DissectMount", create=True) as MockDissectMount,
        ):
            MockTarget.open.return_value = target_bare

            target_bare.volumes.add(Volume(BytesIO(), 1, 0, 0, None, name="first"))
            target_bare.volumes.add(Volume(BytesIO(), 2, 0, 0, None, name="second"))
            target_bare.volumes.add(Volume(BytesIO(), 3, 0, 0, None, name="second_1"))
            target_bare.volumes.add(Volume(BytesIO(), 4, 0, 0, None, name="second"))

            mock_fs = VirtualFilesystem()
            mock_fs.volume = target_bare.volumes[1]
            target_bare.filesystems.add(mock_fs)
            mock_fs = VirtualFilesystem()
            mock_fs.volume = target_bare.volumes[2]
            target_bare.filesystems.add(mock_fs)
            mock_fs = VirtualFilesystem()
            mock_fs.volume = target_bare.volumes[3]
            target_bare.filesystems.add(mock_fs)

            target_mount()

            MockFUSE.assert_called_once()
            MockDissectMount.assert_called_once()
            vfs = MockDissectMount.call_args[0][0]

            assert vfs.listdir("/volumes") == ["first", "second", "second_1", "second_2"]
            assert vfs.listdir("/filesystems") == ["second", "second_1", "second_2"]


def test_mounting_multi_volume_filesystem(target_bare: Target, monkeypatch: pytest.MonkeyPatch) -> None:
    with monkeypatch.context() as m:
        m.setattr("sys.argv", ["target-mount", "mock-target", "mock-mount"])
        m.setattr("dissect.target.tools.mount.HAS_FUSE", True)

        with (
            patch("dissect.target.tools.mount.Target") as MockTarget,
            patch("dissect.target.tools.mount.FUSE", create=True) as MockFUSE,
            patch("dissect.target.tools.mount.DissectMount", create=True) as MockDissectMount,
        ):
            MockTarget.open.return_value = target_bare

            volumes = [
                Volume(BytesIO(), 1, 0, 0, None, name="first"),
                Volume(BytesIO(), 2, 0, 0, None, name="second"),
            ]

            for vol in volumes:
                target_bare.volumes.add(vol)

            mock_fs = VirtualFilesystem()
            mock_fs.volume = volumes
            target_bare.filesystems.add(mock_fs)

            target_mount()

            MockFUSE.assert_called_once()
            MockDissectMount.assert_called_once()
            vfs = MockDissectMount.call_args[0][0]

            assert vfs.listdir("/filesystems") == ["first", "second"]



from __future__ import annotations

import gzip
import json as jsonlib
import lzma
import struct
import subprocess
from configparser import ConfigParser
from configparser import Error as ConfigParserError
from io import BytesIO
from typing import TYPE_CHECKING, Any, BinaryIO, TextIO

from defusedxml import ElementTree
from dissect.hypervisor.util import vmtar
from dissect.sql import sqlite3

from dissect.target.filesystems.nfs import NfsFilesystem
from dissect.target.helpers.sunrpc import client
from dissect.target.helpers.sunrpc.client import LocalPortPolicy

try:
    from dissect.hypervisor.util.envelope import (
        HAS_PYCRYPTODOME,
        HAS_PYSTANDALONE,
        Envelope,
        KeyStore,
    )

    HAS_ENVELOPE = HAS_PYCRYPTODOME or HAS_PYSTANDALONE
except ImportError:
    HAS_ENVELOPE = False

from dissect.target.filesystems import tar
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import OperatingSystem, arg, export, internal
from dissect.target.plugins.os.unix._os import UnixPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem, VirtualFilesystem
    from dissect.target.helpers.fsutil import TargetPath
    from dissect.target.target import Target

VirtualMachineRecord = TargetRecordDescriptor(
    "esxi/vm",
    [
        ("path", "path"),
    ],
)


class ESXiPlugin(UnixPlugin):
    """ESXi OS plugin

    ESXi partitioning varies between versions. Generally, specific partition numbers have special meaning.

    The following is a list of known partition numbers:
        1: EFI boot
        5: BOOTBANK1
        6: BOOTBANK2
        7: vmkcore (ESXi 6), OSDATA / LOCKER (ESXi 7)
        8: store (ESXi 6), HDD VMFS datastore (ESXi 7)
        9: vmkcore (ESXi 6)
    """

    def __init__(self, target: Target):
        super().__init__(target)
        self._config = None
        self._configstore = None

        esx_conf = target.fs.path("/etc/vmware/esx.conf")
        if esx_conf.exists():
            self._config = parse_esx_conf(esx_conf.open("rt", encoding="utf-8"))

        # ESXi 7 introduced the configstore
        # It's made available at /etc/vmware/configstore/current-store-1 during boot, but stored at
        # the path used below in local.tgz
        configstore = target.fs.path("/var/lib/vmware/configstore/backup/current-store-1")
        if configstore.exists():
            self._configstore = parse_config_store(configstore.open())
            self._mount_nfs_shares()
        else:
            self.target.log.warning(
                "No configstore found, some functionality may not work (such as mounting NFS shares)"
            )

    def _cfg(self, path: str) -> str | None:
        if not self._config:
            self.target.log.warning("No ESXi config!")
            return None

        value_name = path.strip("/").split("/")[-1]
        obj = _traverse(path, self._config)

        if not value_name and obj:
            return obj

        return obj.get(value_name) if obj else None

    @classmethod
    def detect(cls, target: Target) -> Filesystem | None:
        bootbanks = [
            fs for fs in target.filesystems if fs.path("boot.cfg").exists() and list(fs.path("/").glob("*.v00"))
        ]

        cfgs = [(fs, parse_boot_cfg(fs.path("boot.cfg").open("rt"))) for fs in bootbanks]
        cfgs.sort(key=lambda pair: pair[1].get("updated", 0), reverse=True)

        return cfgs[0][0] if cfgs else None

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> Self:
        cfg = parse_boot_cfg(sysvol.path("boot.cfg").open("rt"))

        # Mount all the visor tars in individual filesystem layers
        _mount_modules(target, sysvol, cfg)

        # Create a root layer for the "local state" filesystem
        # This stores persistent configuration data
        local_layer = target.fs.append_layer()

        # Mount the local.tgz to the local state layer
        _mount_local(target, local_layer)

        # Mount all other filesystems (VMFS, FAT16)
        _mount_filesystems(target, sysvol, cfg)

        obj = cls(target)

        # Symlink the /var/log directory to the correct destination (if available)
        _link_log_dir(target, cfg, obj)

        return obj

    @export(property=True)
    def hostname(self) -> str:
        if hostname := self._cfg("/adv/Misc/HostName"):
            return hostname.split(".", 1)[0]
        return "localhost"

    @export(property=True)
    def domain(self) -> str | None:
        if hostname := self._cfg("/adv/Misc/HostName"):
            return hostname.partition(".")[2]
        return None

    @export(property=True)
    def ips(self) -> list[str]:
        result = set()
        host_ip = self._cfg("/adv/Misc/HostIPAddr")
        mgmt_ip = self._cfg("/adv/Net/ManagementAddr")

        if host_ip:
            result.add(host_ip)
        if mgmt_ip:
            result.add(mgmt_ip)

        return list(result)

    @export(property=True)
    def version(self) -> str | None:
        boot_cfg = self.target.fs.path("/bootbank/boot.cfg")
        if not boot_cfg.exists():
            return None

        for line in boot_cfg.read_text().splitlines():
            if not line.startswith("build="):
                continue

            _, _, version = line.partition("=")
            return f"VMware ESXi {version.strip()}"
        return None

    @export(record=VirtualMachineRecord)
    def vm_inventory(self) -> Iterator[VirtualMachineRecord]:
        inv_file = self.target.fs.path("/etc/vmware/hostd/vmInventory.xml")
        if not inv_file.exists():
            return []

        root = ElementTree.fromstring(inv_file.read_text("utf-8"))
        for entry in root.iter("ConfigEntry"):
            yield VirtualMachineRecord(
                path=self.target.fs.path(entry.findtext("vmxCfgPath")),
                _target=self.target,
            )

    @export(output="none")
    @arg("path", help="config path")
    @arg("-j", "--json", action="store_true", help="output in JSON format")
    def esxconf(self, path: str, json: bool) -> None:
        obj = self._cfg(path)

        if json:
            print(jsonlib.dumps(obj, indent=4, sort_keys=True))
        else:
            print(obj)

    @internal
    def configstore(self) -> dict[str, Any]:
        return self._configstore

    @export(property=True)
    def os(self) -> str:
        return OperatingSystem.ESXI.value

    def _mount_nfs_shares(self) -> None:
        """Mount NFS shares found in the configstore."""
        if not self._configstore:
            return

        nfs_shares: dict[str, Any] = self._configstore.get("esx", {}).get("storage", {}).get("nfs_v3_datastores", {})
        if not nfs_shares:
            if self._is_nfs_enabled:
                self.target.log.info("No NFS shares found in datastore")
            return

        for key, nfs_share in nfs_shares.items():
            # Parse the NFS share configuration
            user_value: dict[str, Any] = nfs_share.get("user_value", {})
            nfs_ip = user_value.get("hostname", "")
            volume_name = user_value.get("volume_name", "")
            remote_share = user_value.get("remote_share", "")
            if not nfs_ip or not volume_name or not remote_share:
                self.target.log.warning("Invalid NFS share configuration with key: %s", key)
                continue
            mount_point = f"/vmfs/volumes/{volume_name}"
            self._add_nfs(nfs_ip, remote_share, mount_point)

    def _add_nfs(self, nfs_ip: str, remote_share: str, mount_alias: str) -> None:
        """Mount NFS share to the target."""

        if not self._is_nfs_enabled:
            self._log_nfs_mount_disabled(nfs_ip, remote_share, mount_alias)
            return

        uuid = nfs_volume_uuid(nfs_ip, remote_share)
        mount_point = f"/vmfs/volumes/{uuid}"
        try:
            self.target.log.debug("Mounting NFS share %s at %s with alias %s", remote_share, mount_point, mount_alias)

            # On ESXi, there is typically only a single root user.
            # Besides, UnixPlugin::users does not work (see issue https://github.com/fox-it/dissect.target/issues/1093).
            # Moreover, socket rebinding is not implemented on ESXi.
            # Therefore, we try logging only as root. This implies that he NFS share has `no_root_squash` enabled.
            # According to the docs, ESxi mounts NFS shares using root privileges (https://www.vmware.com/docs/vmw-best-practices-running-nfs-vmware-vsphere)  # noqa: E501
            credentials = client.auth_unix("machine", 0, 0, [])
            nfs = NfsFilesystem.connect(nfs_ip, remote_share, credentials, LocalPortPolicy.PRIVILEGED)

            self.target.fs.mount(mount_point, nfs)
            self.target.fs.symlink(mount_point, mount_alias)
        except Exception as e:
            self.target.log.warning(
                "Failed to mount NFS share %s:%s at %s",
                nfs_ip,
                remote_share,
                mount_alias,
            )
            self.target.log.debug("", exc_info=e)


def _mount_modules(target: Target, sysvol: Filesystem, cfg: dict[str, str]) -> None:
    modules = [m.strip() for m in cfg["modules"].split("---")]

    for module in modules:
        module_path = sysvol.path(module)
        if not module_path.exists():
            target.log.warning("Non existent module: %s", module)
            continue

        tfs = None
        if module_path.name.endswith((".tar.gz", ".tgz")):
            tfs = tar.TarFilesystem(module_path.open())
        elif module_path.suffix.startswith(".v"):
            # Visor tar files are always gzipped
            cfile = gzip.GzipFile(fileobj=module_path.open())
            # Sometimes they are also xz compressed, check for XZ magic
            # NOTE: The XZ layer may also contain file signatures
            # Could be interesting to check.
            if cfile.peek(6)[:6] == b"\xfd7zXZ\x00":
                cfile = lzma.LZMAFile(cfile)  # noqa: SIM115

            tfs = tar.TarFilesystem(cfile, tarinfo=vmtar.VisorTarInfo)

        if tfs:
            target.fs.append_layer().mount("/", tfs)


def _mount_local(target: Target, local_layer: VirtualFilesystem) -> None:
    local_tgz = target.fs.path("local.tgz")
    local_tgz_ve = target.fs.path("local.tgz.ve")
    local_fs = None

    if local_tgz.exists():
        local_fs = tar.TarFilesystem(local_tgz.open())
    elif local_tgz_ve.exists():
        # In the case "encryption.info" does not exist, but ".#encryption.info" does
        encryption_info = next(target.fs.path("/").glob("*encryption.info"), None)
        if not local_tgz_ve.exists() or not encryption_info.exists():
            raise ValueError("Unable to find valid configuration archive")

        local_fs = _create_local_fs(target, local_tgz_ve, encryption_info)
    else:
        target.log.warning("No local.tgz or local.tgz.ve found, skipping local state")

    if local_fs:
        local_layer.mount("/", local_fs)


def _decrypt_envelope(local_tgz_ve: TargetPath, encryption_info: TargetPath) -> BinaryIO:
    """Decrypt ``local.tgz.ve`` ourselves with hard-coded keys."""
    envelope = Envelope(local_tgz_ve.open())
    keystore = KeyStore.from_text(encryption_info.read_text("utf-8"))
    return BytesIO(envelope.decrypt(keystore.key, aad=b"ESXConfiguration"))


def _decrypt_crypto_util(local_tgz_ve: TargetPath) -> BytesIO | None:
    """Decrypt ``local.tgz.ve`` using ESXi ``crypto-util``.

    We write to stdout, but this results in ``crypto-util`` exiting with a non-zero return code
    and stderr containing an I/O error message. The file does get properly decrypted, so we return
    ``None`` if there are no bytes in stdout which would indicate it actually failed.
    """

    result = subprocess.run(
        [
            "crypto-util",
            "envelope",
            "extract",
            "--aad",
            "ESXConfiguration",
            f"/{local_tgz_ve.as_posix()}",
            "-",
        ],
        capture_output=True,
    )

    if len(result.stdout) == 0:
        return None

    return BytesIO(result.stdout)


def _create_local_fs(target: Target, local_tgz_ve: TargetPath, encryption_info: TargetPath) -> tar.TarFilesystem | None:
    local_tgz = None

    if HAS_ENVELOPE:
        try:
            local_tgz = _decrypt_envelope(local_tgz_ve, encryption_info)
        except NotImplementedError:
            target.log.debug("Failed to decrypt %s, likely TPM encrypted", local_tgz_ve)
    else:
        target.log.debug("Skipping static decryption because of missing crypto module")

    if not local_tgz and target.name == "local":
        target.log.info(
            "local.tgz is encrypted but static decryption failed, attempting dynamic decryption using crypto-util"
        )
        local_tgz = _decrypt_crypto_util(local_tgz_ve)

        if local_tgz is None:
            target.log.warning("Dynamic decryption of %s failed", local_tgz_ve)
    else:
        target.log.warning("local.tgz is encrypted but static decryption failed and no dynamic decryption available!")

    if local_tgz:
        return tar.TarFilesystem(local_tgz)
    return None


def _mount_filesystems(target: Target, sysvol: Filesystem, cfg: dict[str, str]) -> None:
    version = cfg["build"]

    osdata_fs = None
    locker_fs = None
    for fs in target.filesystems:
        if fs.__type__ == "fat":
            fs.volume.seek(512)
            magic, uuid1, uuid2, uuid3, uuid4 = struct.unpack("<16sIIH6s", fs.volume.read(32))
            if magic != b"VMWARE FAT16    ":
                continue

            fs_uuid = f"{uuid1:08x}-{uuid2:08x}-{uuid3:04x}-{uuid4.hex()}"
            target.fs.mount(f"/vmfs/volumes/{fs_uuid}", fs)
            if not fs.volume.name.startswith("part_"):  # dissect.target quirk
                target.fs.symlink(f"/vmfs/volumes/{fs_uuid}", f"/vmfs/volumes/{fs.volume.name}")

            # ESXi 7 relies on volume names
            # ESXi 6 uses volume numbers
            if fs.volume.name in ("BOOTBANK1", "BOOTBANK2") or (
                fs.volume.number
                in (
                    5,
                    6,
                )
                and fs.exists("boot.cfg")
            ):
                if fs is sysvol:
                    target.fs.symlink(f"/vmfs/volumes/{fs_uuid}", "/bootbank")
                else:
                    target.fs.symlink(f"/vmfs/volumes/{fs_uuid}", "/altbootbank")

            # /store == partition number 8
            if version and version[0] == "6" and fs.volume.number == 8:
                target.fs.symlink(f"/vmfs/volumes/{fs_uuid}", "/store")
                target.fs.symlink("/store", "/locker")

        elif fs.__type__ == "vmfs":
            target.fs.mount(f"/vmfs/volumes/{fs.vmfs.uuid}", fs)
            target.fs.symlink(f"/vmfs/volumes/{fs.vmfs.uuid}", f"/vmfs/volumes/{fs.vmfs.label}")

            if fs.volume.name in ("OSDATA", "LOCKER"):
                target.fs.symlink(
                    f"/vmfs/volumes/{fs.vmfs.uuid}",
                    f"/vmfs/volumes/{fs.volume.name}-{fs.vmfs.uuid}",
                )

                if fs.volume.name == "OSDATA":
                    osdata_fs = fs
                elif fs.volume.name == "LOCKER":
                    locker_fs = fs

    # Symlink /scratch from locker.conf
    # A path _should_ always be set here, even in default installs where it points to OSDATA
    # This file is confusingly called locker.conf
    locker_conf = target.fs.path("/etc/vmware/locker.conf")
    if locker_conf.exists():
        scratch_path = locker_conf.read_text().strip().partition(" ")[0]
        if scratch_path:
            target.fs.symlink(scratch_path, "/scratch")
        else:
            target.log.warning("Scratch path from locker.conf is empty?")
    else:
        target.log.warning("No locker.conf!")

    # Symlink /locker
    # This is marked as legacy in ESXi 7, so may need changing when ESXi 8 comes out
    # Apparently the order is: OSDATA partition -> LOCKER partition -> RAM disk
    # We can't support RAM disk so only check OSDATA and LOCKER
    if osdata_fs:
        target.fs.symlink(f"/vmfs/volumes/OSDATA-{osdata_fs.vmfs.uuid}", "/var/lib/vmware/osdata")
        target.fs.symlink("/var/lib/vmware/osdata/store", "/store")
        target.fs.symlink("/var/lib/vmware/osdata/locker", "/locker")

    elif locker_fs:
        target.fs.symlink(f"/vmfs/volumes/LOCKER-{locker_fs.vmfs.uuid}", "/locker")


def _link_log_dir(target: Target, cfg: dict[str, str], plugin_obj: ESXiPlugin) -> None:
    version = cfg["build"]

    # Don't really know how ESXi does this, but let's just take a shortcut for now
    log_dir = None
    if version and version[0] == "7":
        try:
            log_dir = plugin_obj._configstore["esx"]["syslog"]["global_settings"][""]["user_value"]["log_dir"]
        except KeyError:
            target.log.warning("Failed to read log_dir from configstore, falling back to /scratch/log")
            log_dir = "/scratch/log"
    elif version and version[0] == "6":
        vmsyslog_file = target.fs.path("/etc/vmsyslog.conf")
        if vmsyslog_file.exists():
            try:
                vmsyslog_conf = ConfigParser()
                vmsyslog_conf.read_string(vmsyslog_file.read_text())

                log_dir = vmsyslog_conf.get("vmsyslog", "logdir")
                if log_dir == "<none>":
                    log_dir = "/scratch/log"
            except ConfigParserError as e:
                target.log.warning(
                    "Failed to read log_dir from vmsyslog.conf, falling back to /scratch/log",
                    exc_info=e,
                )
                log_dir = "/scratch/log"

    if log_dir:
        target.fs.symlink(log_dir, "/var/log")


def parse_boot_cfg(fh: TextIO) -> dict[str, str]:
    cfg = {}
    for line in fh:
        line = line.strip()
        if not line or line.startswith("#"):
            continue

        key, _, value = line.partition("=")

        cfg[key.strip()] = value.strip()

    return cfg


def parse_esx_conf(fh: TextIO) -> dict[str, Any]:
    config = {}
    for line in fh:
        line = line.strip()
        if not line:
            continue

        key, _, value = line.partition("=")
        key = key.strip().strip("/")
        value = value.strip().strip('"')

        if value == "true":
            value = True
        elif value == "false":
            value = False
        elif value.isnumeric():
            value = int(value)

        value_name = key.split("/")[-1]
        obj = _traverse(key, config, create=True)
        obj[value_name] = value

    return config


def _traverse(path: str, obj: dict[str, Any], create: bool = False) -> dict[str, Any] | None:
    parts = path.strip("/").split("/")
    path_parts = parts[:-1]
    for part in path_parts:
        array_idx = None
        if part.endswith("]"):
            part, _, rest = part.partition("[")
            array_idx = rest.strip("]")

        if part not in obj:
            if create:
                obj[part] = {}
            else:
                return None

        obj = obj[part]
        if array_idx:
            if array_idx not in obj:
                if create:
                    obj[array_idx] = {}
                else:
                    return None
            obj = obj[array_idx]

    return obj


def parse_config_store(fh: BinaryIO) -> dict[str, Any]:
    db = sqlite3.SQLite3(fh)

    store = {}

    if table := db.table("Config"):
        for row in table.rows():
            component_name = row.Component
            config_group_name = row.ConfigGroup
            value_group_name = row.Name
            identifier_name = row.Identifier

            if component_name not in store:
                store[component_name] = {}
            component = store[component_name]

            if config_group_name not in component:
                component[config_group_name] = {}
            config_group = component[config_group_name]

            if value_group_name not in config_group:
                config_group[value_group_name] = {}
            value_group = config_group[value_group_name]

            if identifier_name not in value_group:
                value_group[identifier_name] = {}
            identifier = value_group[identifier_name]

            identifier["modified_time"] = row.ModifiedTime
            identifier["creation_time"] = row.CreationTime
            identifier["version"] = row.Version
            identifier["success"] = row.Success
            identifier["auto_conf_value"] = jsonlib.loads(row.AutoConfValue) if row.AutoConfValue else None
            identifier["user_value"] = jsonlib.loads(row.UserValue) if row.UserValue else None
            identifier["vital_value"] = jsonlib.loads(row.VitalValue) if row.VitalValue else None
            identifier["cached_value"] = jsonlib.loads(row.CachedValue) if row.CachedValue else None
            identifier["desired_value"] = jsonlib.loads(row.DesiredValue) if row.DesiredValue else None
            identifier["revision"] = row.Revision

    return store


def mix64(a: int, b: int, c: int) -> int:
    """
    Mixes three 64-bit values reversibly.
    """
    # Implement logical right shift by masking first
    a = (a - b - c) ^ ((c & 0xFFFFFFFFFFFFFFFF) >> 43)
    b = (b - c - a) ^ (a << 9)
    c = (c - a - b) ^ ((b & 0xFFFFFFFFFFFFFFFF) >> 8)
    a = (a - b - c) ^ ((c & 0xFFFFFFFFFFFFFFFF) >> 38)
    b = (b - c - a) ^ (a << 23)
    c = (c - a - b) ^ ((b & 0xFFFFFFFFFFFFFFFF) >> 5)
    a = (a - b - c) ^ ((c & 0xFFFFFFFFFFFFFFFF) >> 35)
    b = (b - c - a) ^ (a << 49)
    c = (c - a - b) ^ ((b & 0xFFFFFFFFFFFFFFFF) >> 11)
    a = (a - b - c) ^ ((c & 0xFFFFFFFFFFFFFFFF) >> 12)
    b = (b - c - a) ^ (a << 18)
    c = (c - a - b) ^ ((b & 0xFFFFFFFFFFFFFFFF) >> 22)
    # Normalize to 64 bits
    return a & 0xFFFFFFFFFFFFFFFF, b & 0xFFFFFFFFFFFFFFFF, c & 0xFFFFFFFFFFFFFFFF


def esxi_hash(key: bytes, level: int) -> int:
    """
    Hashes a variable-length key into a 64-bit value.

    This hash function is used in the ESXi kernel.
    It is an exact implementation of the hash3 function defined here: http://burtleburtle.net/bob/c/lookup8.c
    """
    a: int = level
    b: int = level
    c: int = 0x9E3779B97F4A7C13  # Golden ratio, arbitrary value
    bytes_left: int = len(key)
    i: int = 0

    # Process the key in 24-byte chunks
    while bytes_left >= 24:
        a += int.from_bytes(key[i : i + 8], "little")
        b += int.from_bytes(key[i + 8 : i + 16], "little")
        c += int.from_bytes(key[i + 16 : i + 24], "little")
        a, b, c = mix64(a, b, c)
        i += 24
        bytes_left -= 24

    # Handle the last 23 bytes
    c = c + len(key)
    if bytes_left > 0:
        for shift, byte in enumerate(key[i:]):
            if shift < 8:
                a += byte << (shift * 8)
            elif shift < 16:
                b += byte << ((shift - 8) * 8)
            else:
                # c takes 23 - 8 - 8 = 7 bytes (length is added to LSB)
                c += byte << ((shift - 15) * 8)

    _, _, c = mix64(a, b, c)
    return c


def nfs_volume_uuid(host: str, path: str) -> str:
    """Generate a UUID for an NFS volume based on the host and path.

    This is used to create a unique identifier for NFS volumes in ESXi.
    """

    h1 = esxi_hash(host.encode(), 42)  # 42 is starting value
    h2 = esxi_hash(path.encode(), h1)

    low, high = h2 & 0xFFFFFFFF, ((h2 >> 32) & 0xFFFFFFFF)
    return f"{low:8x}-{high:8x}"



#!/usr/bin/env python
from __future__ import annotations

import argparse
import logging
import textwrap

from dissect.target import plugin


def main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--verbose", action="count", default=0, help="increase output verbosity")
    args = parser.parse_args()

    if args.verbose == 1:
        logging.basicConfig(level=logging.ERROR)
    elif args.verbose == 2:
        logging.basicConfig(level=logging.WARNING)
    elif args.verbose == 3:
        logging.basicConfig(level=logging.INFO)
    elif args.verbose >= 4:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.CRITICAL)

    pluginlist = plugin.generate()
    template = """
    from dissect.target.plugin import (
        FailureDescriptor,
        FunctionDescriptor,
        FunctionDescriptorLookup,
        PluginDescriptor,
        PluginDescriptorLookup,
        PluginRegistry,
    )

    PLUGINS = {}
    """
    print(textwrap.dedent(template).format(pluginlist))

    return 0


if __name__ == "__main__":
    main()



from __future__ import annotations

import argparse
from unittest.mock import MagicMock, call, patch

import pytest

from dissect.target.plugin import FailureDescriptor, FunctionDescriptor, PluginRegistry
from dissect.target.target import Event, Target
from dissect.target.tools.report import (
    ExecutionReport,
    TargetExecutionReport,
    format_target_report,
    make_cli_args_overview,
    make_plugin_import_errors_overview,
)


@pytest.fixture
def test_target() -> Target:
    return Target("test_target")


@pytest.fixture
def func_execs() -> set[str]:
    return {"exec2", "exec1"}


@pytest.fixture
def target_execution_report(test_target: Target, func_execs: set[str]) -> TargetExecutionReport:
    return TargetExecutionReport(
        target=test_target,
        func_execs=func_execs,
    )


@pytest.fixture
def incompatible_plugins() -> set[str]:
    return {"incomp_plugin2", "incomp_plugin1"}


@pytest.fixture
def add_incompatible_plugins(target_execution_report: TargetExecutionReport, incompatible_plugins: set[str]) -> None:
    for plugin in incompatible_plugins:
        target_execution_report.add_incompatible_plugin(plugin)


@pytest.fixture
def registered_plugins() -> set[str]:
    return {"regist_plugin2", "regist_plugin1"}


@pytest.fixture
def add_registered_plugins(target_execution_report: TargetExecutionReport, registered_plugins: set[str]) -> None:
    for plugin in registered_plugins:
        target_execution_report.add_registered_plugin(plugin)


@pytest.fixture
def func_errors() -> dict[str, str]:
    return {
        "func1": "trace1",
        "func2": "trace2",
    }


@pytest.fixture
def add_func_errors(target_execution_report: TargetExecutionReport, func_errors: dict[str, str]) -> None:
    for func, trace in func_errors.items():
        target_execution_report.add_func_error(func, trace)


def test_target_execution_report_add_incompatible_plugin(
    target_execution_report: TargetExecutionReport,
    add_incompatible_plugins: None,
    incompatible_plugins: set[str],
) -> None:
    for plugin in target_execution_report.incompatible_plugins:
        assert plugin in incompatible_plugins


def test_target_execution_report_add_registered_plugin(
    target_execution_report: TargetExecutionReport,
    add_registered_plugins: None,
    registered_plugins: set[str],
) -> None:
    for plugin in target_execution_report.registered_plugins:
        assert plugin in registered_plugins


def test_target_execution_report_add_func_error(
    target_execution_report: TargetExecutionReport,
    add_func_errors: None,
    func_errors: dict[str, str],
) -> None:
    for func, trace in target_execution_report.func_errors.items():
        assert func_errors.get(func) == trace


def test_target_execution_report_as_dict(
    test_target: Target,
    target_execution_report: TargetExecutionReport,
    add_incompatible_plugins: None,
    add_registered_plugins: None,
    add_func_errors: None,
    incompatible_plugins: set[str],
    registered_plugins: set[str],
    func_errors: dict[str, str],
    func_execs: set[str],
) -> None:
    report_dict = target_execution_report.as_dict()
    assert report_dict.get("target") == str(test_target)
    assert report_dict.get("incompatible_plugins") == sorted(incompatible_plugins)
    assert report_dict.get("registered_plugins") == sorted(registered_plugins)
    assert report_dict.get("func_errors") == func_errors
    assert report_dict.get("func_execs") == sorted(func_execs)


@pytest.fixture
def execution_report() -> ExecutionReport:
    return ExecutionReport()


@pytest.fixture
def cli_args() -> argparse.Namespace:
    return argparse.Namespace(foo="bar", baz="bla")


@pytest.fixture
def set_cli_args(execution_report: ExecutionReport, cli_args: argparse.Namespace) -> None:
    execution_report.set_cli_args(cli_args)


@pytest.fixture
def plugin_stats() -> PluginRegistry:
    return PluginRegistry(
        __failed__=[
            FailureDescriptor(
                module="plugin1",
                stacktrace="trace1",
            ),
            FailureDescriptor(
                module="plugin2",
                stacktrace="trace2",
            ),
        ]
    )


@pytest.fixture
def set_plugin_stats(execution_report: ExecutionReport, plugin_stats: PluginRegistry) -> None:
    execution_report.set_plugin_stats(plugin_stats)


@pytest.fixture
def target1() -> Target:
    return Target("test1")


@pytest.fixture
def target2() -> Target:
    return Target("test2")


@pytest.fixture
def target_report1(execution_report: ExecutionReport, target1: Target) -> TargetExecutionReport:
    return execution_report.add_target_report(target1)


@pytest.fixture
def target_report2(execution_report: ExecutionReport, target2: Target) -> TargetExecutionReport:
    return execution_report.add_target_report(target2)


@pytest.fixture
def plugin1() -> MagicMock:
    plugin1 = MagicMock()
    plugin1.__module__ = "test_module"
    plugin1.__qualname__ = "plugin1"
    return plugin1


def test_execution_report_set_cli_args(
    execution_report: ExecutionReport,
    set_cli_args: None,
    cli_args: argparse.Namespace,
) -> None:
    assert execution_report.cli_args == vars(cli_args)


def test_execution_report_set_plugin_stats(
    execution_report: ExecutionReport,
    set_plugin_stats: None,
    plugin_stats: PluginRegistry,
) -> None:
    failed_plugins = plugin_stats.__failed__
    assert len(execution_report.plugin_import_errors) == len(failed_plugins)

    for failed_plugin in failed_plugins:
        module = failed_plugin.module
        stacktrace = failed_plugin.stacktrace
        assert execution_report.plugin_import_errors.get(module) == stacktrace


def test_execution_report_get_formatted_report(
    execution_report: ExecutionReport,
    target_report1: TargetExecutionReport,
    target_report2: TargetExecutionReport,
) -> None:
    with (
        patch("dissect.target.tools.report.make_cli_args_overview", return_value="line_1"),
        patch("dissect.target.tools.report.make_plugin_import_errors_overview", return_value="line_2"),
        patch("dissect.target.tools.report.format_target_report", return_value="line_x"),
    ):
        assert execution_report.get_formatted_report() == "line_1\nline_2\nline_x\nline_x"


def test_execution_report_add_target_report(
    execution_report: ExecutionReport,
    target_report1: TargetExecutionReport,
    target_report2: TargetExecutionReport,
) -> None:
    assert len(execution_report.target_reports) == 2
    assert target_report1 in execution_report.target_reports
    assert target_report2 in execution_report.target_reports


def test_execution_report_get_target_report(
    execution_report: ExecutionReport,
    target_report1: TargetExecutionReport,
    target_report2: TargetExecutionReport,
    target1: Target,
    target2: Target,
) -> None:
    assert target_report1 == execution_report.get_target_report(target1)
    assert target_report2 == execution_report.get_target_report(target2)
    target3 = Target("nope")
    assert execution_report.get_target_report(target3) is None
    target_report3 = execution_report.get_target_report(target3, create=True)
    assert target_report3.target == target3


def test_execution_report__get_plugin_name(execution_report: ExecutionReport, plugin1: MagicMock) -> None:
    assert execution_report._get_plugin_name(plugin1) == "test_module.plugin1"


def test_execution_report_log_incompatible_plugin_plugin_cls(
    execution_report: ExecutionReport,
    target_report1: TargetExecutionReport,
    target1: Target,
    plugin1: MagicMock,
) -> None:
    execution_report.log_incompatible_plugin(target1, None, plugin_cls=plugin1)
    assert "test_module.plugin1" in target_report1.incompatible_plugins


def test_execution_report_log_incompatible_plugin_plugin_desc(
    execution_report: ExecutionReport,
    target_report1: TargetExecutionReport,
    target1: Target,
) -> None:
    plugin_desc = FunctionDescriptor(
        name="plugin1",
        namespace=None,
        path="",
        exported=True,
        internal=False,
        findable=True,
        alias=False,
        output=None,
        method_name="plugin1",
        module="test_module",
        qualname="plugin1",
    )
    execution_report.log_incompatible_plugin(target1, None, plugin_desc=plugin_desc)
    assert "test_module.plugin1" in target_report1.incompatible_plugins


def test_execution_report_log_registered_plugin(
    execution_report: ExecutionReport,
    target_report1: TargetExecutionReport,
    target1: Target,
) -> None:
    execution_report.log_registered_plugin(target1, None, plugin_inst=MagicMock())
    assert "unittest.mock.MagicMock" in target_report1.registered_plugins


def test_execution_report_log_func_error(
    execution_report: ExecutionReport,
    target_report1: TargetExecutionReport,
    target1: Target,
    func_errors: dict[str, str],
) -> None:
    func, trace = next(iter(func_errors.items()))
    execution_report.log_func_error(target1, None, func, trace)
    assert target_report1.func_errors.get(func) == trace


def test_execution_report_log_func_execution(
    execution_report: ExecutionReport,
    target_report1: TargetExecutionReport,
    target1: Target,
    func_execs: set[str],
) -> None:
    func = next(iter(func_execs))
    execution_report.log_func_execution(target1, None, func)
    assert func in target_report1.func_execs


def test_execution_report_set_event_callbacks(execution_report: ExecutionReport) -> None:
    mock_target = MagicMock()
    event_callbacks = (
        (Event.INCOMPATIBLE_PLUGIN, execution_report.log_incompatible_plugin),
        (Event.REGISTERED_PLUGIN, execution_report.log_registered_plugin),
        (Event.FUNC_EXEC, execution_report.log_func_execution),
        (Event.FUNC_EXEC_ERROR, execution_report.log_func_error),
    )

    execution_report.set_event_callbacks(mock_target)

    assert len(mock_target.mock_calls) == len(event_callbacks)
    for event_type, event_callback in event_callbacks:
        mock_call = call.set_event_callback(event_type=event_type, event_callback=event_callback)
        assert mock_call in mock_target.mock_calls


def test_execution_report_as_dict(
    execution_report: ExecutionReport,
    set_plugin_stats: None,
    plugin_stats: PluginRegistry,
    target_report1: TargetExecutionReport,
    target_report2: TargetExecutionReport,
    set_cli_args: None,
    cli_args: argparse.Namespace,
) -> None:
    expected_dict = {
        "plugin_import_errors": {
            "plugin1": "trace1",
            "plugin2": "trace2",
        },
        "target_reports": ["report1", "report2"],
        "cli_args": {
            "foo": "bar",
            "baz": "bla",
        },
    }

    with (
        patch.object(target_report1, "as_dict", return_value="report1"),
        patch.object(target_report2, "as_dict", return_value="report2"),
    ):
        execution_report_dict = execution_report.as_dict()
        assert execution_report_dict == expected_dict


def test_report_make_cli_args_overview(
    execution_report: ExecutionReport,
    set_cli_args: None,
    cli_args: argparse.Namespace,
) -> None:
    cli_args_overview = make_cli_args_overview(execution_report)
    assert "foo: bar" in cli_args_overview
    assert "baz: bla" in cli_args_overview


def test_report_make_plugin_import_errors_overview(
    execution_report: ExecutionReport,
    set_plugin_stats: None,
    plugin_stats: PluginRegistry,
) -> None:
    plugin_import_errors_overview = make_plugin_import_errors_overview(execution_report)
    assert "plugin1:\n        trace1" in plugin_import_errors_overview
    assert "plugin2:\n        trace2" in plugin_import_errors_overview


def test_report_format_target_report(
    test_target: Target,
    target_execution_report: TargetExecutionReport,
    add_incompatible_plugins: None,
    add_registered_plugins: None,
    add_func_errors: None,
    incompatible_plugins: set[str],
    registered_plugins: set[str],
    func_errors: dict[str, str],
    func_execs: set[str],
) -> None:
    target_report = format_target_report(target_execution_report)
    assert str(test_target) in target_report

    for registered_plugin in registered_plugins:
        assert registered_plugin in target_report

    for incompatible_plugin in incompatible_plugins:
        assert incompatible_plugin in target_report

    for func, trace in func_errors.items():
        assert func in target_report
        assert trace in target_report

    for func_exec in func_execs:
        assert func_exec in target_report



from __future__ import annotations

import hashlib
import logging
from io import BytesIO
from typing import BinaryIO

from dissect.cstruct import cstruct

from dissect.target.plugins.os.windows.dpapi.crypto import (
    CipherAlgorithm,
    HashAlgorithm,
    derive_password_hash,
    dpapi_hmac,
)

try:
    from Crypto.Hash import MD4

    HAS_CRYPTO = True
except ImportError:
    HAS_CRYPTO = False

log = logging.getLogger(__name__)


master_key_def = """
struct DomainKey {
    DWORD   dwVersion;
    DWORD   secretLen;
    DWORD   accessCheckLen;
    char    guid[16];
    char    encryptedSecret[secretLen];
    char    accessCheck[accessCheckLen];
};

struct CredHist {
    DWORD   dwVersion;
    char    guid[16];
};

struct MasterKey {
    DWORD   dwVersion;
    char    pSalt[16];
    DWORD   dwPBKDF2IterationCount;
    DWORD   HMACAlgId;                  // This is actually ALG_ID
    DWORD   CryptAlgId;                 // This is actually ALG_ID
    // BYTE    pKey[];
};

struct CredSystem {
    DWORD   dwRevision;
    char    pMachine[20];
    char    pUser[20];
};

struct MasterKeyFileHeader {
    DWORD   dwVersion;                  // Masterkey version. Should be 1 or 2
    DWORD   dwReserved1;
    DWORD   dwReserved2;
    WCHAR   szGuid[36];                 // GUID of master key. Should match filename
    DWORD   dwUnused1;
    DWORD   dwUnused2;
    DWORD   dwPolicy;
    QWORD   qwUserKeySize;
    QWORD   qwLocalEncKeySize;
    QWORD   qwLocalKeySize;
    QWORD   qwDomainKeySize;
};
"""
c_master_key = cstruct().load(master_key_def)


class MasterKey:
    def __init__(self, data: bytes):
        buf = BytesIO(data)
        self._mk = c_master_key.MasterKey(buf)
        self._mk_key = buf.read()

        self.key = None
        self.key_hash = None
        self.decrypted = False

    def decrypt_with_hash(self, user_sid: str, password_hash: bytes) -> bool:
        """Decrypts the master key with the given user's SID and password hash."""
        return self.decrypt_with_key(derive_password_hash(password_hash, user_sid))

    def decrypt_with_hash_10(self, user_sid: str, password_hash: bytes) -> bool:
        """Decrypts the master key with the given user's hash and SID.

        Newer version of :meth:`~MasterKey.decrypt_with_hash`
        """
        user_sid_encoded = user_sid.encode("utf-16-le")
        pwd_hash1 = hashlib.pbkdf2_hmac("sha256", password_hash, user_sid_encoded, 10000)
        pwd_hash2 = hashlib.pbkdf2_hmac("sha256", pwd_hash1, user_sid_encoded, 1)[0:16]
        return self.decrypt_with_key(derive_password_hash(pwd_hash2, user_sid))

    def decrypt_with_password(self, user_sid: str, pwd: str) -> bool:
        """Decrypts the master key with the given user's password and SID."""
        if self.decrypted:
            return True

        pwd = pwd.encode("utf-16-le")

        for algo in ["sha1", "md4"]:
            if algo in hashlib.algorithms_available:
                pwd_hash = hashlib.new(algo, pwd)
            elif HAS_CRYPTO and algo == "md4":
                pwd_hash = MD4.new(pwd)
            else:
                log.warning("No cryptography capabilities for algorithm %s", algo)
                continue

            self.decrypt_with_key(derive_password_hash(pwd_hash.digest(), user_sid))
            if self.decrypted:
                break

        return self.decrypted

    def decrypt_with_key(self, key: bytes) -> bool:
        """Decrypts the master key with the given encryption key.

        This function also extracts the HMAC part of the decrypted data and compares it with the computed one.

        Note that once successfully decrypted, this function turns into a no-op.
        """
        if self.decrypted:
            return True

        if not self._mk_key:
            return False

        # Compute encryption key
        hash_algo = HashAlgorithm.from_id(self._mk.HMACAlgId)
        cipher_algo = CipherAlgorithm.from_id(self._mk.CryptAlgId)

        data = cipher_algo.decrypt_with_hmac(
            self._mk_key,
            key,
            self._mk.pSalt,
            hash_algo,
            self._mk.dwPBKDF2IterationCount,
        )

        self.key = data[-64:]
        self.hmac_salt = data[:16]
        self.hmac = data[16 : 16 + int(hash_algo.digest_length)]
        self.hmac_computed = dpapi_hmac(key, self.hmac_salt, self.key, hash_algo)
        self.decrypted = self.hmac == self.hmac_computed
        if self.decrypted:
            self.key_hash = hashlib.sha1(self.key).digest()

        return self.decrypted


class MasterKeyFile:
    def __init__(self, fh: BinaryIO):
        self._mk_header = c_master_key.MasterKeyFileHeader(fh)
        self._user_mk = None

        # User Master Key
        if self._mk_header.qwUserKeySize:
            self._user_mk = MasterKey(fh.read(self._mk_header.qwUserKeySize))

        # Here we would also parse the rest of the keys, but as of now we don't decrypt them
        self._backup_mk = None
        self._credhist_mk = None
        self._domain_mk = None

    @property
    def decrypted(self) -> bool:
        return self._user_mk.decrypted

    @property
    def key(self) -> bytes:
        return self._user_mk.key

    def decrypt_with_hash(self, user_sid: str, password_hash: bytes) -> bool:
        """See :meth:`MasterKey.decrypt_with_hash` and :meth:`MasterKey.decrypt_with_hash_10`."""
        if not self._user_mk.decrypted:
            self._user_mk.decrypt_with_hash_10(user_sid, password_hash)

        if not self._user_mk.decrypted:
            self._user_mk.decrypt_with_hash(user_sid, password_hash)

        return self._user_mk.decrypted

    def decrypt_with_password(self, user_sid: str, pwd: str) -> bool:
        """See :meth:`MasterKey.decrypt_with_password`."""
        return self._user_mk.decrypt_with_password(user_sid, pwd)

    def decrypt_with_key(self, key: bytes) -> bool:
        """See :meth:`MasterKey.decrypt_with_key`."""
        return self._user_mk.decrypt_with_key(key)


class CredSystem:
    def __init__(self, buf: bytes):
        self._struct = c_master_key.CredSystem(buf)
        self.machine_key = self._struct.pMachine
        self.user_key = self._struct.pUser



from __future__ import annotations

import base64
import textwrap
from enum import Enum, auto
from io import BytesIO
from typing import TYPE_CHECKING

import pytest

from dissect.target.plugins.apps.ssh.openssh import (
    OpenSSHPlugin,
    calculate_fingerprints,
)

if TYPE_CHECKING:
    from typing_extensions import Self

    from dissect.target.filesystem import VirtualFilesystem
    from dissect.target.target import Target


@pytest.fixture(
    params=[
        ("target_unix_users", "fs_unix"),
        ("target_win_users", "fs_win"),
    ],
    ids=["target_unix", "target_windows"],
)
def target_and_filesystem(request: pytest.FixtureRequest) -> tuple[Target, VirtualFilesystem]:
    target = request.getfixturevalue(request.param[0])
    filesystem = request.getfixturevalue(request.param[1])
    return target, filesystem


class TargetDir(Enum):
    HOME = auto()
    SSHD = auto()


class Alternatives:
    user_name: str
    home_dir: str
    sshd_dir: str
    os_type: str
    separator: str
    label: str

    @classmethod
    def from_target(cls, target: Target) -> Self:
        alternative = cls()
        alternative.os_type = target._os.os
        alternative.separator = "/"
        alternative.label = "NO_LABEL"

        if alternative.os_type == "windows":
            alternative.separator = target.fs.alt_separator
            alternative.home_dir = "C:\\Users\\John"
            alternative.sshd_dir = "C:\\ProgramData\\ssh"
            alternative.user_name = "John"
            alternative.label = "C:\\"
        else:
            alternative.home_dir = "/root"
            alternative.sshd_dir = "/etc/ssh"
            alternative.user_name = "root"

        return alternative

    def mapping_path(self, file_name: str, target_dir: TargetDir) -> str:
        return self.filesystem_path(file_name, target_dir).strip(self.label)

    def filesystem_path(self, file_name: str, target_dir: TargetDir) -> str:
        mapping = {TargetDir.HOME: self.home_dir, TargetDir.SSHD: self.sshd_dir}
        return self.separator.join([mapping.get(target_dir), *file_name.split("/")])


def test_authorized_keys_plugin(target_and_filesystem: tuple[Target, VirtualFilesystem]) -> None:
    target, fs = target_and_filesystem
    authorized_keys_data = """
    ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAINz6oq+IweAoQFMzQ0aJLYXJFkLn3tXMbVZ550wvUKOw
    ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAINz6oq+IweAoQFMzQ0aJLYXJFkLn3tXMbVZ550wvUKOw comment
    ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAINz6oq+IweAoQFMzQ0aJLYXJFkLn3tXMbVZ550wvUKOw long comment with spaces
    command="foo bar" ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAINz6oq+IweAoQFMzQ0aJLYXJFkLn3tXMbVZ550wvUKOw
    command="foo bar" ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAINz6oq+IweAoQFMzQ0aJLYXJFkLn3tXMbVZ550wvUKOw with a comment
    # Invalid
    ssh-ed25519
    """

    target_system = Alternatives.from_target(target)

    fs.map_file_fh(
        target_system.mapping_path(".ssh/authorized_keys", target_dir=TargetDir.HOME),
        BytesIO(textwrap.dedent(authorized_keys_data).encode()),
    )
    fs.map_file_fh(
        target_system.mapping_path("administrator_authorized_keys", target_dir=TargetDir.SSHD),
        BytesIO(textwrap.dedent(authorized_keys_data).encode()),
    )

    plugin = OpenSSHPlugin(target)

    results = list(plugin.authorized_keys())
    assert len(results) == 10

    assert results[0].key_type == "ssh-ed25519"
    assert results[0].options is None
    assert results[0].comment == ""

    assert results[1].key_type == "ssh-ed25519"
    assert results[1].options is None
    assert results[1].comment == "comment"

    assert results[2].key_type == "ssh-ed25519"
    assert results[2].options is None
    assert results[2].comment == "long comment with spaces"

    assert results[3].key_type == "ssh-ed25519"
    assert results[3].options == 'command="foo bar"'
    assert results[3].comment == ""

    assert results[4].key_type == "ssh-ed25519"
    assert results[4].options == 'command="foo bar"'
    assert results[4].comment == "with a comment"


def test_known_hosts_plugin(target_and_filesystem: tuple[Target, VirtualFilesystem]) -> None:
    target, fs = target_and_filesystem

    known_hosts_data = """
    # Comments allowed at start of line
    cvs.example.net,192.0.2.10 ssh-rsa AAAA1234.....= comment with spaces
    # A hashed hostname
    |1|JfKTdBh7rNbXkVAQCRp4OQoPfmI=|USECr3SWf1JUPsms5AqfD5QfxkM= ssh-rsa AAAA1234.....=
    # A revoked key
    @revoked * ssh-rsa AAAAB5W...
    # A CA key, accepted for any host in *.mydomain.com or *.mydomain.org
    @cert-authority *.mydomain.org,*.mydomain.com ssh-rsa AAAAB5W...
    """

    target_system = Alternatives.from_target(target)

    fs.map_file_fh(
        target_system.mapping_path(".ssh/known_hosts", target_dir=TargetDir.HOME),
        BytesIO(textwrap.dedent(known_hosts_data).encode()),
    )

    target.add_plugin(OpenSSHPlugin)

    results = list(target.openssh.known_hosts())
    assert len(results) == 6

    assert results[0].host == "cvs.example.net"
    assert results[0].key_type == "ssh-rsa"
    assert results[0].comment == "comment with spaces"
    assert results[0].marker is None

    assert results[1].host == "192.0.2.10"
    assert results[1].key_type == "ssh-rsa"
    assert results[1].comment == "comment with spaces"
    assert results[1].marker is None

    assert results[2].host == "|1|JfKTdBh7rNbXkVAQCRp4OQoPfmI=|USECr3SWf1JUPsms5AqfD5QfxkM="
    assert results[2].key_type == "ssh-rsa"
    assert results[2].comment == ""
    assert results[2].marker is None

    assert results[3].host == "*"
    assert results[3].key_type == "ssh-rsa"
    assert results[3].comment == ""
    assert results[3].marker == "@revoked"

    assert results[4].host == "*.mydomain.org"
    assert results[4].key_type == "ssh-rsa"
    assert results[4].comment == ""
    assert results[4].marker == "@cert-authority"

    assert results[5].host == "*.mydomain.com"
    assert results[5].key_type == "ssh-rsa"
    assert results[5].comment == ""
    assert results[5].marker == "@cert-authority"
    assert results[5].username == target_system.user_name


def test_private_keys_plugin_rfc4716_ed25519(target_and_filesystem: tuple[Target, VirtualFilesystem]) -> None:
    target, fs = target_and_filesystem
    private_key_data = """
    -----BEGIN OPENSSH PRIVATE KEY-----
    b3BlbnNzaC1rZXktdjEAAAAABG5vbmUAAAAEbm9uZQAAAAAAAAABAAAAMwAAAAtzc2gtZW
    QyNTUxOQAAACDc+qKviMHgKEBTM0NGiS2FyRZC597VzG1WeedML1CjsAAAAJhebt8UXm7f
    FAAAAAtzc2gtZWQyNTUxOQAAACDc+qKviMHgKEBTM0NGiS2FyRZC597VzG1WeedML1CjsA
    AAAECwfJEGqm+M4uYAWj3RMwA/6OcwmM48QT2MTCU5XSZpmdz6oq+IweAoQFMzQ0aJLYXJ
    FkLn3tXMbVZ550wvUKOwAAAAEWxvbmcgY29tbWVudCBoZXJlAQIDBA==
    -----END OPENSSH PRIVATE KEY-----
    """

    target_system = Alternatives.from_target(target)

    public_key_data = "AAAAC3NzaC1lZDI1NTE5AAAAINz6oq+IweAoQFMzQ0aJLYXJFkLn3tXMbVZ550wvUKOw"

    fs.map_file_fh(
        target_system.mapping_path("ssh_host_ed25519_key", TargetDir.SSHD),
        BytesIO(textwrap.dedent(private_key_data).encode()),
    )

    fs.map_file_fh(
        target_system.mapping_path(".ssh/garbage_file", TargetDir.HOME),
        BytesIO(textwrap.dedent("FOO PRIVATE KEY----- BAR").encode()),
    )

    target.add_plugin(OpenSSHPlugin)

    results = list(target.openssh.private_keys())
    assert len(results) == 1

    private_key = results[0]
    assert private_key.key_format == "RFC4716"
    assert private_key.key_type == "ssh-ed25519"
    assert private_key.comment == "long comment here"
    assert private_key.public_key == public_key_data
    assert not private_key.encrypted
    assert str(private_key.path).replace("\\", "/") == target_system.filesystem_path(
        "ssh_host_ed25519_key", TargetDir.SSHD
    ).replace(target_system.label, "\\sysvol\\").replace("\\", "/")


def test_private_keys_plugin_rfc4716_rsa_encrypted(target_and_filesystem: tuple[Target, VirtualFilesystem]) -> None:
    target, fs = target_and_filesystem
    # Generated using password "password".
    private_key_data = """
    -----BEGIN OPENSSH PRIVATE KEY-----
    b3BlbnNzaC1rZXktdjEAAAAACmFlczI1Ni1jdHIAAAAGYmNyeXB0AAAAGAAAABCC2yy1YK
    QYzL52ycXms2QZAAAAEAAAAAEAAAGXAAAAB3NzaC1yc2EAAAADAQABAAABgQDBz7gMDWRd
    6VMfxUEXLwYVfk4Tt7C9luYtoFcU6GcdUx8lhhZX4tBCTth4Abm1EDyBX+OFt0UO9tHaYI
    dctzn0tAaEIoV6vpGEOx3NyqwG63DFruuJVYaQKfnMqADMBTjKGXbSKMosJA9bLT+mR8oG
    ObnadUfi3/n4QmmKtyXt+THxwDvM4qTplEsgss912LtOVGBfUj2a1HLljTlydz/0JlC20O
    nG516XLq7aM/TRXvYWxa5jWRwlsP795SA2Zd9aeAk6d5PvYRIcdPnmW9DwN3LXFq/7JNMH
    VpB/reHlBbTUOGbbKP8Ba71fgcYoSPDaXtkUsjaHjEFpBpwjROcJKUxiCPcrHyBxaKpI8Q
    aeJWXyBaIUWJbsBFy65F8EOPZYR95ZI4yQvmw2lp75g7DwNCp7Xu3jljaeIzR2LDLk282x
    0A++Di6tG+IzVvMOcgEJIzykofOEXMJt0oVhnvIr9y5VWl/G8X1e20GqjiImLSGKgvZnzP
    x93RmjHXfOQk8AAAWA8mTnTOARPA69ZLPq1wOTiPQftt5O+bRWuxGhux4E1jVsTA+lqB6c
    nEKFjquQLD78YWMrJ85ypxweMzF+9Yyx0LLTLY1FZYPYNQV0xli3wEaWh7M5s18anXiJby
    +gvicjPL4DFKqVvCQ3w7z+HbMLjfe+OSZlwoGoM2d/HefdIZSGXfByYD+R+d5Arai6XQbt
    GWe2+rJWMoZlXGXTvLvxrUn2drioW+9ewq1iw7E2JN5bz8TKv6y+ExP6OX0YwPrxVyvf1u
    70U0KZv87rP8dOvtWTEWP2vNOgQB9WoMoH9BICVwOsQgCXTKnMQpcH30KEFnHlCcXivt99
    fjvXgbyJ8l6fELbE7Rrq8r9ZiH0ICtO6DQ5SPfhkp3IN3F3ZC+qo7sVDzG79rXQfWU9DCi
    q8SFCCTIg38uyFhSK+HUmpg3oMpLQ2HKjkpXjVNLlPqojGBpZUQUeCkFm9vGqQUkl5KYlx
    xz3tpqa9bfSG+jEIYURBDBIEuChzv+6CIO8wP8oQlLCB2ZX97bWKQgSqubJ6qcl7WNpyj8
    LBfJLsmgraNoIp/+Z2at8+RIkAnOGimhSS6Mx926coPbZJiapIVnxtN/8M6kpqKLQa4yoo
    NPyZWlKVRddFXdsFZiW7CagMqVfQtiAtVl54Wn/BrUzHGtP9QlHXYojvoevz3M7YhhnaqH
    INJ1oKZKtFACCbb7+KnbFONv03aXmKkmnVDtBEPmumqIdctl8tV06lOhoI6rX/ryh7lYSj
    hVC7L24adLAPL7YCueRcxY3dx4QPL4jWOPXC8jvDD+SEHn9F+sOf3ao6Opp4eJQYU7gkdL
    Wr8dt+JZSbhvw846H+mDm7ZuyXMRezoEfxx335lsUmdwLxhUcmpJN6RjUQb/mKWOoSX5iX
    FBsLQDboJGrMvMv6bY3tFZB3EYfJcaBzYLrL/PA0orLj7rekgs14D4KhjqXgIwL3U9s/Am
    pdoTggrhLkUGleyC8gYrjei+BNqtopfF11kUiSqA4OrO2Li3SG7BpHryTZQgRF4LtB2lre
    cudqfa7/ShW1sr+76SWvXMa4ikOPUwcuA0CIKmdZ5Xdo4/WNOqxLTOOOMOB6BRwY4KfO64
    WVVak2mfl7CX1baei718liH9ho/Me7MgulHSXZMcziHr+wmjK21PjgcH36L07pYM606XLi
    UxEcLBfWhaytHNk/Zogy+0hXohbFj7YTj6Ccmn+BeH+hZs8T1cCuZhd61HJqYxMifxNvWj
    1/1cn9uIRuyxaP925TDPBLG1j/Vu9LaBUmzfDx2LfZl96NlSox3Ja43s2vgdUvWMrht4Zv
    KfvvoChYZMHSIeRVcp/yrclSvQWCWIvvPXtwLUCxTM2KB7ZzYaqj882oiAcTh+Sz0l/+4R
    Bnl+nCzarY+Sf0wjdq6SsM4GDDc/ll80H8Yqn7WGj2z4QdJ2KCOGI9QgbaU319Au8sdimt
    jqlQrBNwm8j89ZqDpcjmC3ZE/dqZ9h8uSZmf5zkBb9OfHZcoemrGg+ZSBMlYaRVRk5ddB2
    dI9wxbhVvrnn0D2BTHesnoyETeCezx0MLNCA2ZXzQhoGDsRdN3Y2fO6B2vG+fGW4niBb4K
    URzv/rrxllZNn/Xnn43Bl4m7R/cOfxnX0UKP3e1WWT+E3hqQkY8fdQbS3m6r6ckqBDOIwP
    nVib6pd+kouoGtyWgPJD8rDmFh3/yv5P+Mn8LzlUd1HEXYzQTe8S/YYnEHPdBl+Q5aLwhn
    ffTBVvsbm3B/uec4lp+NTzvzRN2LK4euNKl2/Kd6pmPC5GcIcf2esKFdV4sgcSNxlupFPe
    4F5ifByPNQCED7+37LU19H3jpoOpabVxdR5y1zaDFqvI4D/VSfEzrVbNrWEJUrD/tCkr1B
    vp9YaQ==
    -----END OPENSSH PRIVATE KEY-----"""

    public_key_data = (
        "AAAAB3NzaC1yc2EAAAADAQABAAABgQDBz7gMDWRd6VMfxUEXLwYVfk4Tt7C9luYtoFcU6GcdUx8lhhZX4tBCTth4Abm1EDyBX+OFt0UO9tHaYI"
        "dctzn0tAaEIoV6vpGEOx3NyqwG63DFruuJVYaQKfnMqADMBTjKGXbSKMosJA9bLT+mR8oGObnadUfi3/n4QmmKtyXt+THxwDvM4qTplEsgss91"
        "2LtOVGBfUj2a1HLljTlydz/0JlC20OnG516XLq7aM/TRXvYWxa5jWRwlsP795SA2Zd9aeAk6d5PvYRIcdPnmW9DwN3LXFq/7JNMHVpB/reHlBb"
        "TUOGbbKP8Ba71fgcYoSPDaXtkUsjaHjEFpBpwjROcJKUxiCPcrHyBxaKpI8QaeJWXyBaIUWJbsBFy65F8EOPZYR95ZI4yQvmw2lp75g7DwNCp7"
        "Xu3jljaeIzR2LDLk282x0A++Di6tG+IzVvMOcgEJIzykofOEXMJt0oVhnvIr9y5VWl/G8X1e20GqjiImLSGKgvZnzPx93RmjHXfOQk8="
    )

    target_system = Alternatives.from_target(target)

    fs.map_file_fh(
        target_system.mapping_path(".ssh/id_rsa", TargetDir.HOME),
        BytesIO(textwrap.dedent(private_key_data).encode()),
    )

    target.add_plugin(OpenSSHPlugin)

    results = list(target.openssh.private_keys())
    private_key = results[0]

    assert len(results) == 1
    assert private_key.key_format == "RFC4716"
    assert private_key.key_type == "ssh-rsa"
    assert private_key.public_key == public_key_data
    assert private_key.comment == ""
    assert private_key.encrypted
    assert str(private_key.path).replace("\\", "/") == target_system.filesystem_path(
        ".ssh/id_rsa", TargetDir.HOME
    ).replace("\\", "/")


def test_private_keys_plugin_pem_ecdsa(target_and_filesystem: tuple[Target, VirtualFilesystem]) -> None:
    target, fs = target_and_filesystem
    private_key_data = """
    -----BEGIN EC PRIVATE KEY-----
    MHcCAQEEIPLzo1QWDkkMxKEG3X8fW7495i5GtVeQhnkpvXMJTZAooAoGCCqGSM49
    AwEHoUQDQgAEn14FkGUBSSX0ocobnqqcWpeSPKgUJDj1sZtXwLTdfPkZ4++J/HiA
    T9q+r/zG2BunHjtTaa0isjJgXWw++3JAzg==
    -----END EC PRIVATE KEY-----"""

    target_system = Alternatives.from_target(target)

    fs.map_file_fh(
        target_system.mapping_path(".ssh/id_ecdsa", TargetDir.HOME),
        BytesIO(textwrap.dedent(private_key_data).encode()),
    )

    target.add_plugin(OpenSSHPlugin)

    results = list(target.openssh.private_keys())
    private_key = results[0]

    assert len(results) == 1
    assert private_key.key_format == "PEM"
    assert private_key.key_type == "ecdsa"
    assert private_key.username == target_system.user_name
    assert not private_key.encrypted
    assert str(private_key.path).replace("\\", "/") == target_system.filesystem_path(
        ".ssh/id_ecdsa", TargetDir.HOME
    ).replace("\\", "/")


def test_private_keys_plugin_pkcs8_dsa(target_and_filesystem: tuple[Target, VirtualFilesystem]) -> None:
    target, fs = target_and_filesystem
    private_key_data = """
    -----BEGIN PRIVATE KEY-----
    MIIBSgIBADCCASsGByqGSM44BAEwggEeAoGBAP78NtXw6e2YgD3caU3LbY3fxCtz
    W9R+AF4s3nlbAKCx9lVtnSnEE+sbByfinV3iCmlf8muU0AS/7E9aFMqkDM5sG+cK
    ttJ689Ef/RmrZP2QE8YyZgKLGiK2HVmPVvBEhfWA/Pge45zuPoEAf33czqC6LF0k
    Kq0842BuGsx4iYx3AhUAph7ywYR16EZyIKpQvVZjEoqgp3sCgYBbjazw8sAjAEFX
    xpPl58Iy52wMb53/8tpOUBU1Nn4jPUmgzszrd1cpt4TGtqHasi2uWWyyPmYQe/i2
    YfMY4MAosSA9ZaqIi4Dkd3L+7rIS4zMy7LcTuhaex0n4V0qYBjRZJFXV9iqBN83l
    Klm96maH1SL9x5+GZR6mAjYyfZ+4sgQWAhROqDuRIc5ilMqjV0pG2spo7HVuwg==
    -----END PRIVATE KEY-----"""

    target_system = Alternatives.from_target(target)

    fs.map_file_fh(
        target_system.mapping_path(".ssh/id_dsa", TargetDir.HOME),
        BytesIO(textwrap.dedent(private_key_data).encode()),
    )

    target.add_plugin(OpenSSHPlugin)

    results = list(target.openssh.private_keys())
    private_key = results[0]

    assert len(results) == 1
    assert private_key.key_format == "PKCS8"
    assert private_key.username == target_system.user_name
    assert not private_key.encrypted
    assert str(private_key.path).replace("\\", "/") == target_system.filesystem_path(
        ".ssh/id_dsa", TargetDir.HOME
    ).replace("\\", "/")


def test_public_keys_plugin(target_and_filesystem: tuple[Target, VirtualFilesystem]) -> None:
    target, fs = target_and_filesystem
    user_public_key_data = (
        "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAINz6oq+IweAoQFMzQ0aJLYXJFkLn3tXMbVZ550wvUKOw long comment here"
    )

    host_public_key_data = (
        "ssh-rsa AAAAB3NzaC1yc2EAAAADAQABAAABgQDUZo/RgdRqnitmlWcD7dXcFuIr5528xaJ65exm7FCqrQDMsdlV8naH6bxSmmA0eLUS/YUFwr"
        "YRdkC925rRE37kNfleDycbm6j5QCAT6eZZgftUEUicpcqA52PYzzzGi1g56n3DkkxkoKA5XZkTilHEZLLV1zgUR9GzzvvT8VpgwyNCm9qxfsBC"
        "tywrg6MCqBY/0ssh156pDVmHHr7eWUBkdlj4I16YHLQRPblt9jnUx7S37T/ZRFTRtupePcGCEU9X4D5fWmRfJMIyNtd6LJ9wAXh/h2D+tKW6Th"
        "A1rFDrnYQqcUR2zjSkfSCyYGkt9CnoYhpFdEv3fYRuNCUM8hKa01mGtfy4Vnwh5TL6M4uKhdUR4DZ5CLesciTEz4zbmeXhY5vIij0vsXaP047E"
        "QhwLXXTyyqZ+Pt5oCfrFBtDMtv9NxBMsZ1XGYuSlrc1mqasFBeg0SpIADsJ5DMaWUkkuoeF414PYd8Emu+B6In3h6aHEsP7ELkr9C+qImzIly4"
        "s= comment"
    )

    target_system = Alternatives.from_target(target)

    fs.map_file_fh(
        target_system.mapping_path(".ssh/id_ed25519.pub", TargetDir.HOME),
        BytesIO(textwrap.dedent(user_public_key_data).encode()),
    )

    fs.map_file_fh(
        target_system.mapping_path("ssh_host_rsa_key.pub", TargetDir.SSHD),
        BytesIO(textwrap.dedent(host_public_key_data).encode()),
    )

    target.add_plugin(OpenSSHPlugin)

    results = list(target.openssh.public_keys())

    assert len(results) == 2

    user_public_key = results[0]
    host_public_key = results[1]

    assert user_public_key.key_type == user_public_key_data.split(" ", 2)[0]
    assert user_public_key.public_key == user_public_key_data.split(" ", 2)[1]
    assert user_public_key.comment == user_public_key_data.split(" ", 2)[2]
    assert user_public_key.fingerprint.md5 == "1f3d475966231eeb5455c8485dd030e4"
    assert user_public_key.fingerprint.sha1 == "e39242ca1d74bea99285b212e908e18cc67e4dec"
    assert user_public_key.fingerprint.sha256 == "7b77007b0b51a86ced6b5fe25639092484c4c39cf76b283ef65fdf49a00f44d2"
    assert str(user_public_key.path).replace("\\", "/") == target_system.filesystem_path(
        ".ssh/id_ed25519.pub", TargetDir.HOME
    ).replace("\\", "/")

    assert host_public_key.key_type == host_public_key_data.split(" ", 2)[0]
    assert host_public_key.public_key == host_public_key_data.split(" ", 2)[1]
    assert host_public_key.comment == host_public_key_data.split(" ", 2)[2]
    assert host_public_key.fingerprint.md5 == "a3f2ebfa8d16efd321015e1618fd281b"
    assert host_public_key.fingerprint.sha1 == "f6656cc642fb08f53a1df77d0acff9852a649989"
    assert host_public_key.fingerprint.sha256 == "8c7023d563c763fcf5104332d7cf51c978c5ba1dd9f5cbd341edd32dfcbef3ef"
    assert str(host_public_key.path).replace("\\", "/") == target_system.filesystem_path(
        "ssh_host_rsa_key.pub", TargetDir.SSHD
    ).replace(target_system.label, "\\sysvol\\").replace("\\", "/")


def test_calculate_fingerprints() -> None:
    ed25519_pub = "ssh-ed25519 AAAAC3NzaC1lZDI1NTE5AAAAINz6oq+IweAoQFMzQ0aJLYXJFkLn3tXMbVZ550wvUKOw long comment here"

    assert calculate_fingerprints(base64.b64decode(ed25519_pub.split(" ")[1])) == (
        "1f3d475966231eeb5455c8485dd030e4",
        "e39242ca1d74bea99285b212e908e18cc67e4dec",
        "7b77007b0b51a86ced6b5fe25639092484c4c39cf76b283ef65fdf49a00f44d2",
    )

    assert calculate_fingerprints(base64.b64decode(ed25519_pub.split(" ")[1]), ssh_keygen_format=True) == (
        "1f3d475966231eeb5455c8485dd030e4",
        "45JCyh10vqmShbIS6QjhjMZ+Tew",
        "e3cAewtRqGzta1/iVjkJJITEw5z3ayg+9l/fSaAPRNI",
    )



from __future__ import annotations

import itertools
import re
from datetime import datetime
from functools import cached_property
from typing import TYPE_CHECKING, NamedTuple

from dissect.target.exceptions import FileNotFoundError, UnsupportedPluginError
from dissect.target.helpers.fsutil import open_decompress
from dissect.target.plugin import OperatingSystem, export
from dissect.target.plugins.apps.webserver.webserver import (
    WebserverAccessLogRecord,
    WebserverErrorLogRecord,
    WebserverHostRecord,
    WebserverPlugin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.target import Target


class LogFormat(NamedTuple):
    name: str
    pattern: re.Pattern


# e.g. ServerRoot "/etc/httpd"
RE_CONFIG_ROOT = re.compile(
    r"""
        [\s#]*                          # Optionally prefixed by space(s) or pound sign(s).
        ServerRoot
        \s
        "?(?P<location>[^"\s]+)"
        $
    """,
    re.VERBOSE | re.IGNORECASE,
)

# e.g. Include conf.modules.d/*.conf and IncludeOptional conf.d/*.conf
RE_CONFIG_INCLUDE = re.compile(
    r"""
        [\s#]*                          # Optionally prefixed by space(s) or pound sign(s).
        (Include|IncludeOptional)       # Directive indicating that additional config files are loaded.
        \s
        ?(?P<location>[^"\s]+)
        $
    """,
    re.VERBOSE | re.IGNORECASE,
)


# e.g. CustomLog "/custom/log/location/access.log" common
RE_CONFIG_CUSTOM_LOG_DIRECTIVE = re.compile(
    r"""
        [\s#]*                          # Optionally prefixed by space(s) or pound sign(s).
        CustomLog                       # Directive indicating that a custom access log location / format is used.
        \s
        "?(?P<location>[^"\s]+)"?       # Location to log to, optionally wrapped in double quotes.
        \s
        (?P<logformat>[^$]+)            # Format to use (can be either a format string or a nickname).
        $
    """,
    re.VERBOSE | re.IGNORECASE,
)

# e.g ErrorLog "/var/log/httpd/error_log"
RE_CONFIG_ERRORLOG_DIRECTIVE = re.compile(
    r"""
        [\s#]*                          # Optionally prefixed by space(s) or pound sign(s).
        ErrorLog                        # Directive indicating that a custom error log location / format is used.
        \s
        "?(?P<location>[^"\s]+)"?       # Location to log to, optionally wrapped in double quotes.
        $
    """,
    re.VERBOSE | re.IGNORECASE,
)

RE_REMOTE_PATTERN = r"""
    (?P<remote_ip>.*?)                  # Client IP address of the request.
    \s
    (?P<remote_logname>.*?)             # Remote logname (from identd, if supplied).
    \s
    (?P<remote_user>.*?)                # Remote user if the request was authenticated.
"""

RE_REFERER_USER_AGENT_PATTERN = r"""
    "(?P<referer>.*?)"                  # Value of the 'Referer' HTTP Header.
    \s
    "(?P<useragent>.*?)"                # Value of the 'User-Agent' HTTP Header.
"""

RE_RESPONSE_TIME_PATTERN = r"""
(
    "
    Time:\s
    (?P<response_time>.*?)              # Time taken to serve the response, including a unit of measurement.
    "
)
"""

RE_ACCESS_COMMON_PATTERN = r"""
    \[(?P<ts>[^\]]*)\]                  # Timestamp including milliseconds.
    \s
    (\[(?P<pid>[0-9]+)\]\s)?            # The process ID of the child that serviced the request (optional).
    "
    (
        -                               # Malformed requests may result in the value "-"
        |
        (
            (?P<method>.*?)             # The HTTP Method used for the request.
            \s
            (?P<uri>.*?)                # The HTTP URI of the request.
            \s
            ?(?P<protocol>HTTP\/.*?)?   # The request protocol.
        )
    )
    "
    \s
    (?P<status_code>\d{3})              # The HTTP Status Code of the response.
    \s
    (?P<bytes_sent>-|\d+)               # Bytes sent, including headers.
"""

RE_ERROR_COMMON_PATTERN = r"""
    \[
        (?P<ts>[^\]]*)                  # Timestamp including milliseconds.
    \]
    \s
    \[
        (?P<module>[^:]*)               # Name of the module logging the message.
        \:
        (?P<level>[^]]*)                # Loglevel of the message.
    \]
    \s
    \[
        pid\s(?P<pid>\d*)               # Process ID of current process.
        (\:tid\s(?P<tid>\d*))?          # Thread ID of current thread (optional).
    \]
    \s
    ((?P<error_source>[^\:]*)\:\s)?     # Source file name and line number of the log call (optional).
    (
        \[
            client\s(?P<client>[^]]+)   # Client IP address and port of the request (optional).
        \]\s
    )?
    ((?P<error_code>\w+)\:\s)?          # APR/OS error status code and string (optional).
    (?P<message>.*)                     # The actual log message.
"""

RE_ENV_VAR_IN_STRING = re.compile(r"\$\{(?P<env_var>[^\"\s$]+)\}", re.VERBOSE)

RE_VIRTUALHOST = re.compile(r"^\<VirtualHost (?P<addr>[^\s:]+)(?:\:(?P<port>\d+))?", re.IGNORECASE)

LOG_FORMAT_ACCESS_COMMON = LogFormat(
    "common",
    re.compile(
        rf"{RE_REMOTE_PATTERN}\s{RE_ACCESS_COMMON_PATTERN}",
        re.VERBOSE,
    ),
)
LOG_FORMAT_ACCESS_VHOST_COMBINED = LogFormat(
    "vhost_combined",
    re.compile(
        rf"""
        (?P<server_name>.*?):(?P<port>.*)
        \s
        {RE_REMOTE_PATTERN}
        \s
        {RE_ACCESS_COMMON_PATTERN}
        \s
        {RE_REFERER_USER_AGENT_PATTERN}
        """,
        re.VERBOSE,
    ),
)
LOG_FORMAT_ACCESS_COMBINED = LogFormat(
    "combined",
    re.compile(
        rf"{RE_REMOTE_PATTERN}\s{RE_ACCESS_COMMON_PATTERN}\s{RE_REFERER_USER_AGENT_PATTERN}",
        re.VERBOSE,
    ),
)
LOG_FORMAT_ERROR_COMMON = LogFormat("error", re.compile(RE_ERROR_COMMON_PATTERN, re.VERBOSE))


def apache_response_time_to_ms(time_str: str) -> int:
    """Convert a string containing amount and measurement (e.g. '10000 microsecs') to milliseconds."""
    amount, _, measurement = time_str.partition(" ")
    amount = int(amount)
    if measurement == "microsecs":
        return amount // 1000
    raise ValueError(f"Could not parse {time_str}")


class ApachePlugin(WebserverPlugin):
    """Apache log parsing plugin.

    Apache has three default access log formats, which this plugin can all parse automatically. These are::

        LogFormat "%v:%p %h %l %u %t \\"%r\\" %>s %O \\"%{Referer}i\\" \\"%{User-Agent}i\\"" vhost_combined
        LogFormat "%h %l %u %t \\"%r\\" %>s %O \\"%{Referer}i\\" \\"%{User-Agent}i\\"" combined
        LogFormat "%h %l %u %t \\"%r\\" %>s %O" common

    For the definitions of each format string, see https://httpd.apache.org/docs/2.4/mod/mod_log_config.html#formats

    For Apache, the error logs by default follow the following format::

        ErrorLogFormat "[%{u}t] [%-m:%l] [pid %P:tid %T] %7F: %E: [client\\ %a] %M% ,\\ referer\\ %{Referer}i"
    """

    __namespace__ = "apache"

    DEFAULT_LOG_DIRS = (
        "/var/log/apache2",
        "/var/log/apache",
        "/var/log/httpd",
        "/var/log",
        "sysvol/xampp/apache/logs",
        "/opt/lampp/logs",
    )
    ACCESS_LOG_NAMES = ("access.log", "access_log", "httpd-access.log")
    ERROR_LOG_NAMES = ("error.log",)
    DEFAULT_CONFIG_PATHS = (
        "/etc/apache2/apache2.conf",
        "/usr/local/etc/apache22/httpd.conf",
        "/usr/local/apache2/httpd.conf",
        "/etc/httpd/conf/httpd.conf",
        "/etc/httpd.conf",
    )
    DEFAULT_ENVVAR_PATHS = ("/etc/apache2/envvars", "/etc/sysconfig/httpd", "/etc/rc.conf")
    DEFAULT_SERVER_ROOTS = ("/etc/apache2", "/usr/local/apache2", "/etc/httpd", "/home/httpd", "/home/apache2")

    def __init__(self, target: Target):
        super().__init__(target)
        self.server_root = None
        self.access_paths = set()
        self.error_paths = set()
        self.virtual_hosts = set()
        self.find_logs()

    def check_compatible(self) -> None:
        if not self.access_paths and not self.error_paths and not self.virtual_hosts:
            raise UnsupportedPluginError("No Apache log files or virtual hosts found")

        if self.target.os == OperatingSystem.CITRIX:
            raise UnsupportedPluginError("Use the 'apps.webserver.citrix' apache plugin instead")

    def find_logs(self) -> None:
        """Discover any present Apache log paths on the target system.
        Populates ``self.access_paths``, ``self.error_paths`` and ``self.virtual_hosts``.

        References:
            - https://httpd.apache.org/docs/2.4/logs.html
            - https://httpd.apache.org/docs/2.4/mod/mod_log_config.html
            - https://www.cyberciti.biz/faq/apache-logs/
            - https://unix.stackexchange.com/a/269090
        """

        # Check if any well known default Apache log locations exist
        for log_dir, log_name in itertools.product(self.DEFAULT_LOG_DIRS, self.ACCESS_LOG_NAMES):
            self.access_paths.update(self.target.fs.path(log_dir).glob(f"*{log_name}*"))

        for log_dir, log_name in itertools.product(self.DEFAULT_LOG_DIRS, self.ERROR_LOG_NAMES):
            self.error_paths.update(self.target.fs.path(log_dir).glob(f"*{log_name}*"))

        seen = set()

        # Check default Apache configs for CustomLog or ErrorLog directives
        for config in self.DEFAULT_CONFIG_PATHS:
            if (path := self.target.fs.path(config)).exists() and path not in seen:
                self._process_conf_file(path, seen)

        # Check all .conf files inside the server root
        if self.server_root:
            for path in self.server_root.rglob("*.conf"):
                if path not in seen:
                    self._process_conf_file(path, seen)

    def _process_conf_file(self, path: Path, seen: set[Path] | None = None) -> None:
        """Process an Apache ``.conf`` file for ``ServerRoot``, ``CustomLog``, ``Include``
        and ``OptionalInclude`` directives. Populates ``self.access_paths`` and ``self.error_paths``.

        Apache / Httpd directives are case-insensitive which is why we convert lines to lowercase for pattern matching.

        References:
            - https://httpd.apache.org/docs/2.4/en/configuring.html#syntax
        """
        seen = set() if seen is None else seen

        if path in seen:
            self.target.log.warning("Detected recursion in Apache configuration, file already parsed: %s", path)
            return

        seen.add(path)

        for line in path.open("rt"):
            if not (line := line.strip()):
                continue

            line_lower = line.lower()

            if "serverroot" in line_lower:
                if not (match := RE_CONFIG_ROOT.match(line)):
                    self.target.log.warning("Unable to parse Apache 'ServerRoot' configuration in %s: %r", path, line)
                    continue
                location = match.groupdict().get("location")
                self.server_root = self.target.fs.path(location)

            elif "customlog" in line_lower or "errorlog" in line_lower:
                self._process_conf_line(path, line)

            elif "include" in line_lower:
                if not (match := RE_CONFIG_INCLUDE.match(line)):
                    self.target.log.warning("Unable to parse Apache 'Include' configuration in %s: %r", path, line)
                    continue

                location = match.groupdict().get("location")

                if "*" in location:
                    root, rest = location.split("*", 1)
                    if root.startswith("/"):
                        root = self.target.fs.path(root)
                    elif not root.startswith("/") and self.server_root:
                        root = self.server_root.joinpath(root)
                    elif not self.server_root:
                        self.target.log.warning("Unable to resolve relative Include in %s: %r", path, line)
                        continue

                    for found_conf in root.glob(f"*{rest}"):
                        self._process_conf_file(found_conf, seen)

                elif (
                    # Relative from server root
                    (self.server_root and (include_path := self.server_root.joinpath(location)).exists())
                    or
                    # Absolute path
                    ((include_path := self.target.fs.path(location)).exists())
                ):
                    self._process_conf_file(include_path, seen)

                else:
                    self.target.log.warning("Unable to resolve Apache Include in %s: %r", path, line)

            # While we're at it, see if we can find any VirtualHosts
            elif "<virtualhost" in line_lower:
                self.virtual_hosts.add(path)

    def _process_conf_line(self, path: Path, line: str) -> None:
        """Parse and resolve the given ``CustomLog`` or ``ErrorLog`` directive found in a Apache ``.conf`` file."""
        if "errorlog" in line.lower():
            pattern = RE_CONFIG_ERRORLOG_DIRECTIVE
            set_to_update = self.error_paths
        else:
            pattern = RE_CONFIG_CUSTOM_LOG_DIRECTIVE
            set_to_update = self.access_paths

        if not (match := pattern.match(line)):
            self.target.log.warning("Unexpected Apache 'ErrorLog' or 'CustomLog' configuration in %s: %r", path, line)
            return

        location = match.groupdict().get("location")
        custom_log = self.target.fs.path(location)

        if env_var_match := RE_ENV_VAR_IN_STRING.match(location):
            env_var = env_var_match.groupdict()["env_var"]
            if (apache_log_dir := self.env_vars.get(env_var)) is None:
                self.target.log.warning("%s does not exist, cannot resolve '%s' in %s", env_var, custom_log, path)
                return

            custom_log = self.target.fs.path(location.replace(f"${{{env_var}}}", apache_log_dir.replace("$SUFFIX", "")))

        set_to_update.update(path for path in custom_log.parent.glob(f"*{custom_log.name}*"))

    @cached_property
    def env_vars(self) -> dict[str, str]:
        variables = {}
        for envvar_file in self.DEFAULT_ENVVAR_PATHS:
            if (file := self.target.fs.path(envvar_file)).exists():
                for line in file.read_text().splitlines():
                    key, _, value = line.strip().partition("=")
                    if key:
                        variables[key.replace("export ", "")] = value.strip("\"'")
        return variables

    @export(record=WebserverAccessLogRecord)
    def access(self) -> Iterator[WebserverAccessLogRecord]:
        """Return contents of Apache access log files in unified ``WebserverAccessLogRecord`` format."""
        for line, path in self._iterate_log_lines(self.access_paths):
            try:
                logformat = self.infer_access_log_format(line)
                if not logformat:
                    self.target.log.warning("Apache log format could not be inferred for log line: %s (%s)", line, path)
                    continue

                match = logformat.pattern.match(line)
                if not match:
                    self.target.log.warning(
                        "Could not match Apache log format %s for log line: %s (%s)", logformat.name, line, path
                    )
                    continue

                log = match.groupdict()
                if response_time := log.get("response_time"):
                    response_time = apache_response_time_to_ms(response_time)

                yield WebserverAccessLogRecord(
                    ts=datetime.strptime(log["ts"], "%d/%b/%Y:%H:%M:%S %z"),
                    remote_user=clean_value(log["remote_user"]),
                    remote_ip=log["remote_ip"],
                    local_ip=clean_value(log.get("local_ip")),
                    method=log["method"],
                    uri=log["uri"],
                    protocol=log["protocol"],
                    status_code=log["status_code"],
                    bytes_sent=clean_value(log["bytes_sent"]) or 0,
                    pid=log.get("pid"),
                    referer=clean_value(log.get("referer")),
                    useragent=clean_value(log.get("useragent")),
                    response_time_ms=response_time,
                    source=path,
                    _target=self.target,
                )

            except Exception as e:
                self.target.log.warning("An error occured parsing Apache log file %s: %s", path, str(e))
                self.target.log.debug("", exc_info=e)

    @export(record=WebserverErrorLogRecord)
    def error(self) -> Iterator[WebserverErrorLogRecord]:
        """Return contents of Apache error log files in unified ``WebserverErrorLogRecord`` format."""
        for line, path in self._iterate_log_lines(self.error_paths):
            try:
                match = LOG_FORMAT_ERROR_COMMON.pattern.match(line)
                if not match:
                    self.target.log.warning("Could not match Apache error log format for log line: %s (%s)", line, path)
                    continue

                log = match.groupdict()
                remote_ip = log.get("client")
                if remote_ip and ":" in remote_ip:
                    remote_ip, _, port = remote_ip.rpartition(":")
                error_source = log.get("error_source")
                error_code = log.get("error_code")

                # Both error_source and error_code follow the same logformat. When both are present, the error source
                # goes before the client and the error code goes after. However, it is also possible that only the error
                # code is available, in which case it is situated *after* the client. In such situations our regex match
                # has assigned the variables wrong, and we need to do a swap.
                if error_source and error_code is None:
                    error_source, error_code = error_code, error_source

                # Unlike with access logs, ErrorLogFormat doesn't log the offset to UTC but insteads logs in local time.
                ts = self.target.datetime.local(datetime.strptime(log["ts"], "%a %b %d %H:%M:%S.%f %Y"))  # noqa: DTZ007

                yield WebserverErrorLogRecord(
                    ts=ts,
                    pid=log.get("pid"),
                    remote_ip=remote_ip,
                    module=log["module"],
                    level=log["level"],
                    error_source=error_source,
                    error_code=error_code,
                    message=log["message"],
                    source=path,
                    _target=self.target,
                )

            except Exception as e:
                self.target.log.warning("An error occured parsing Apache log file %s: %s", path, str(e))
                self.target.log.debug("", exc_info=e)

    @export(record=WebserverHostRecord)
    def hosts(self) -> Iterator[WebserverHostRecord]:
        """Return found ``VirtualHost`` directives in the Apache configuration.

        References:
            - https://httpd.apache.org/docs/2.4/mod/core.html#virtualhost
        """

        for path in self.virtual_hosts:
            # A configuration file can contain multiple VirtualHost directives.
            current_vhost = {}
            for line in path.open("rt"):
                line_lower = line.lower()
                if "<virtualhost" in line_lower:
                    # Currently only supports a single addr:port combination.
                    if match := RE_VIRTUALHOST.match(line.lstrip()):
                        current_vhost = match.groupdict()
                    else:
                        self.target.log.warning("Unable to parse VirtualHost directive %r in %s", line, path)
                        current_vhost = {}

                elif "</virtualhost" in line_lower:
                    yield WebserverHostRecord(
                        ts=path.lstat().st_mtime,
                        server_name=current_vhost.get("servername") or current_vhost.get("addr"),
                        server_port=current_vhost.get("port"),
                        root_path=current_vhost.get("documentroot"),
                        access_log_config=current_vhost.get("customlog", "").rpartition(" ")[0],
                        error_log_config=current_vhost.get("errorlog"),
                        source=path,
                        _target=self.target,
                    )

                else:
                    key, _, value = line.strip().partition(" ")
                    current_vhost[key.lower()] = value

    def _iterate_log_lines(self, paths: list[Path]) -> Iterator[tuple[str, Path]]:
        """Iterate through a list of paths and yield tuples of loglines and the path of the file where they're from."""
        for path in paths:
            try:
                path = path.resolve(strict=True)
                for line in open_decompress(path, "rt"):
                    line = line.strip()
                    if not line:
                        continue
                    yield line, path
            except FileNotFoundError:
                self.target.log.warning("Apache log file configured but could not be found (dead symlink?): %s", path)

    @staticmethod
    def infer_access_log_format(line: str) -> LogFormat | None:
        """Attempt to infer what standard LogFormat is used. Returns None if no known format can be inferred.

        Three default log type examples from Apache (note that the ipv4 could also be ipv6)

        Combined::

            1.2.3.4 - - [19/Dec/2022:17:25:12 +0100] "GET / HTTP/1.1" 304 247 "-" "Mozilla/5.0
                        (Windows NT 10.0; Win64; x64); AppleWebKit/537.36 (KHTML, like Gecko) Chrome/108.0.0.0
                        Safari/537.36\"

        Common::

            1.2.3.4 - - [19/Dec/2022:17:25:40 +0100] "GET / HTTP/1.1" 200 312
            1.2.3.4 - - [19/Dec/2022:17:25:40 +0100] "GET / HTTP/1.1" 301 -

        vhost_combined::

            example.com:80 1.2.3.4 - - [19/Dec/2022:17:25:40 +0100] "GET / HTTP/1.1" 200 312 "-"
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64); AppleWebKit/537.36 (KHTML, like Gecko)
            Chrome/108.0.0.0 Safari/537.36\"
        """
        parts = line.split()
        first_part = parts[0]
        if ":" in first_part and "." in first_part:
            # does not start with IP, hence it must be a vhost typed log
            return LOG_FORMAT_ACCESS_VHOST_COMBINED
        if line[-1] == '"':
            # ends with a quotation mark but does not contain a response time, meaning there is only a user agent
            return LOG_FORMAT_ACCESS_COMBINED
        if line[-1].isdigit() or line[-1] == "-":
            # ends with a digit or '-' indicating response size in bytes
            return LOG_FORMAT_ACCESS_COMMON

        return None


def clean_value(value: str | None) -> str | None:
    """Clean the given value by replacing empty strings and ``"-"`` with ``None``."""

    if value in ("-", ""):
        return None

    return value
